# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_log import log as logging
from oslo_utils import importutils
from oslo_utils import strutils
import webob.exc

import trove.common.apischema as apischema
from trove.common import cfg
from trove.common.db.mysql import models as guest_models
from trove.common import exception
from trove.common.i18n import _
from trove.common import notification
from trove.common.notification import StartNotification
from trove.common import pagination
from trove.common.utils import correct_id_with_req
from trove.common import wsgi
from trove.extensions.common.service import ExtensionController
from trove.extensions.mysql.common import populate_users
from trove.extensions.mysql.common import populate_validated_databases
from trove.extensions.mysql.common import unquote_user_host
from trove.extensions.mysql import models
from trove.extensions.mysql import views


LOG = logging.getLogger(__name__)
import_class = importutils.import_class
CONF = cfg.CONF


class UserController(ExtensionController):
    """Controller for instance functionality."""
    schemas = apischema.user

    @classmethod
    def get_schema(cls, action, body):
        action_schema = super(UserController, cls).get_schema(action, body)
        if 'update_all' == action:
            update_type = list(body.keys())[0]
            action_schema = action_schema.get(update_type, {})
        return action_schema

    def index(self, req, tenant_id, instance_id):
        """Return all users."""
        LOG.info("Listing users for instance '%(id)s'\n"
                 "req : '%(req)s'\n\n",
                 {"id": instance_id, "req": req})
        context = req.environ[wsgi.CONTEXT_KEY]
        self.authorize_target_action(context, 'user:index', instance_id)
        users, next_marker = models.Users.load(context, instance_id)
        view = views.UsersView(users)
        paged = pagination.SimplePaginatedDataView(req.url, 'users', view,
                                                   next_marker)
        return wsgi.Result(paged.data(), 200)

    def create(self, req, body, tenant_id, instance_id):
        """Creates a set of users."""
        LOG.info("Creating users for instance '%(id)s'\n"
                 "req : '%(req)s'\n\n"
                 "body: '%(body)s'\n'n",
                 {"id": instance_id,
                  "req": strutils.mask_password(req),
                  "body": strutils.mask_password(body)})
        context = req.environ[wsgi.CONTEXT_KEY]
        self.authorize_target_action(context, 'user:create', instance_id)
        context.notification = notification.DBaaSUserCreate(context,
                                                            request=req)
        users = body['users']
        with StartNotification(context, instance_id=instance_id,
                               username=",".join([user['name']
                                                  for user in users])):
            try:
                model_users = populate_users(users)
                models.User.create(context, instance_id, model_users)
            except (ValueError, AttributeError) as e:
                raise exception.BadRequest(_("User create error: %(e)s")
                                           % {'e': e})
        return wsgi.Result(None, 202)

    def delete(self, req, tenant_id, instance_id, id):
        LOG.info("Delete instance '%(id)s'\n"
                 "req : '%(req)s'\n\n",
                 {"id": instance_id, "req": req})
        context = req.environ[wsgi.CONTEXT_KEY]
        self.authorize_target_action(context, 'user:delete', instance_id)
        id = correct_id_with_req(id, req)
        username, host = unquote_user_host(id)
        user = None
        context.notification = notification.DBaaSUserDelete(context,
                                                            request=req)
        with StartNotification(context, instance_id=instance_id,
                               username=username):
            try:
                user = guest_models.MySQLUser(name=username,
                                              host=host)
                found_user = models.User.load(context, instance_id, username,
                                              host)
                if not found_user:
                    user = None
            except (ValueError, AttributeError) as e:
                raise exception.BadRequest(_("User delete error: %(e)s")
                                           % {'e': e})
            if not user:
                raise exception.UserNotFound(uuid=id)
            models.User.delete(context, instance_id, user.serialize())
        return wsgi.Result(None, 202)

    def show(self, req, tenant_id, instance_id, id):
        """Return a single user."""
        LOG.info("Showing a user for instance '%(id)s'\n"
                 "req : '%(req)s'\n\n",
                 {"id": instance_id, "req": req})
        context = req.environ[wsgi.CONTEXT_KEY]
        self.authorize_target_action(context, 'user:show', instance_id)
        id = correct_id_with_req(id, req)
        username, host = unquote_user_host(id)
        user = None
        try:
            user = models.User.load(context, instance_id, username, host)
        except (ValueError, AttributeError) as e:
            raise exception.BadRequest(_("User show error: %(e)s")
                                       % {'e': e})
        if not user:
            raise exception.UserNotFound(uuid=id)
        view = views.UserView(user)
        return wsgi.Result(view.data(), 200)

    def update(self, req, body, tenant_id, instance_id, id):
        """Change attributes for one user."""
        LOG.info("Updating user attributes for instance '%(id)s'\n"
                 "req : '%(req)s'\n\n",
                 {"id": instance_id, "req": strutils.mask_password(req)})
        context = req.environ[wsgi.CONTEXT_KEY]
        self.authorize_target_action(context, 'user:update', instance_id)
        id = correct_id_with_req(id, req)
        username, hostname = unquote_user_host(id)
        user = None
        user_attrs = body['user']
        context.notification = notification.DBaaSUserUpdateAttributes(
            context, request=req)
        with StartNotification(context, instance_id=instance_id,
                               username=username):
            try:
                user = models.User.load(context, instance_id, username,
                                        hostname)
            except (ValueError, AttributeError) as e:
                raise exception.BadRequest(_("Error loading user: %(e)s")
                                           % {'e': e})
            if not user:
                raise exception.UserNotFound(uuid=id)
            try:
                models.User.update_attributes(context, instance_id, username,
                                              hostname, user_attrs)
            except (ValueError, AttributeError) as e:
                raise exception.BadRequest(_("User update error: %(e)s")
                                           % {'e': e})
        return wsgi.Result(None, 202)

    def update_all(self, req, body, tenant_id, instance_id):
        """Change the password of one or more users."""
        LOG.info("Updating user password for instance '%(id)s'\n"
                 "req : '%(req)s'\n\n",
                 {"id": instance_id, "req": strutils.mask_password(req)})
        context = req.environ[wsgi.CONTEXT_KEY]
        self.authorize_target_action(context, 'user:update_all', instance_id)
        context.notification = notification.DBaaSUserChangePassword(
            context, request=req)
        users = body['users']
        model_users = []
        with StartNotification(context, instance_id=instance_id,
                               username=",".join([user['name']
                                                  for user in users])):
            for user in users:
                try:
                    mu = guest_models.MySQLUser(name=user['name'],
                                                host=user.get('host'),
                                                password=user['password'])
                    found_user = models.User.load(context, instance_id,
                                                  mu.name, mu.host)
                    if not found_user:
                        user_and_host = mu.name
                        if mu.host:
                            user_and_host += '@' + mu.host
                        raise exception.UserNotFound(uuid=user_and_host)
                    model_users.append(mu)
                except (ValueError, AttributeError) as e:
                    raise exception.BadRequest(_("Error loading user: %(e)s")
                                               % {'e': e})
            try:
                models.User.change_password(context, instance_id, model_users)
            except (ValueError, AttributeError) as e:
                raise exception.BadRequest(_("User password update error: "
                                             "%(e)s")
                                           % {'e': e})
        return wsgi.Result(None, 202)


class UserAccessController(ExtensionController):
    """Controller for adding and removing database access for a user."""
    schemas = apischema.user

    @classmethod
    def get_schema(cls, action, body):
        schema = {}
        if 'update_all' == action:
            schema = cls.schemas.get(action).get('databases')
        return schema

    def _get_user(self, context, instance_id, user_id):
        username, hostname = unquote_user_host(user_id)
        try:
            user = models.User.load(context, instance_id, username, hostname)
        except (ValueError, AttributeError) as e:
            raise exception.BadRequest(_("Error loading user: %(e)s")
                                       % {'e': e})
        if not user:
            raise exception.UserNotFound(uuid=user_id)
        return user

    def index(self, req, tenant_id, instance_id, user_id):
        """Show permissions for the given user."""
        LOG.info("Showing user access for instance '%(id)s'\n"
                 "req : '%(req)s'\n\n",
                 {"id": instance_id, "req": req})

        context = req.environ[wsgi.CONTEXT_KEY]
        self.authorize_target_action(
            context, 'user_access:index', instance_id)
        # Make sure this user exists.
        user_id = correct_id_with_req(user_id, req)
        user = self._get_user(context, instance_id, user_id)
        if not user:
            LOG.error("No such user: %(user)s ", {'user': user})
            raise exception.UserNotFound(uuid=user)
        username, hostname = unquote_user_host(user_id)
        access = models.User.access(context, instance_id, username, hostname)
        view = views.UserAccessView(access.databases)
        return wsgi.Result(view.data(), 200)

    def update(self, req, body, tenant_id, instance_id, user_id):
        """Grant access for a user to one or more databases."""
        LOG.info("Granting user access for instance '%(id)s'\n"
                 "req : '%(req)s'\n\n",
                 {"id": instance_id, "req": req})
        context = req.environ[wsgi.CONTEXT_KEY]
        self.authorize_target_action(
            context, 'user_access:update', instance_id)
        context.notification = notification.DBaaSUserGrant(
            context, request=req)
        user_id = correct_id_with_req(user_id, req)
        user = self._get_user(context, instance_id, user_id)
        if not user:
            LOG.error("No such user: %(user)s ", {'user': user})
            raise exception.UserNotFound(uuid=user)
        username, hostname = unquote_user_host(user_id)
        databases = [db['name'] for db in body['databases']]
        with StartNotification(context, instance_id=instance_id,
                               username=username, database=databases):
            models.User.grant(context, instance_id, username, hostname,
                              databases)
        return wsgi.Result(None, 202)

    def delete(self, req, tenant_id, instance_id, user_id, id):
        """Revoke access for a user."""
        LOG.info("Revoking user access for instance '%(id)s'\n"
                 "req : '%(req)s'\n\n",
                 {"id": instance_id, "req": req})
        context = req.environ[wsgi.CONTEXT_KEY]
        self.authorize_target_action(
            context, 'user_access:delete', instance_id)
        context.notification = notification.DBaaSUserRevoke(
            context, request=req)
        user_id = correct_id_with_req(user_id, req)
        user = self._get_user(context, instance_id, user_id)
        if not user:
            LOG.error("No such user: %(user)s ", {'user': user})
            raise exception.UserNotFound(uuid=user)
        username, hostname = unquote_user_host(user_id)
        access = models.User.access(context, instance_id, username, hostname)
        databases = [db.name for db in access.databases]
        with StartNotification(context, instance_id=instance_id,
                               username=username, database=databases):
            if id not in databases:
                raise exception.DatabaseNotFound(uuid=id)
            models.User.revoke(context, instance_id, username, hostname, id)
        return wsgi.Result(None, 202)


class SchemaController(ExtensionController):
    """Controller for instance functionality."""
    schemas = apischema.dbschema

    def index(self, req, tenant_id, instance_id):
        """Return all schemas."""
        LOG.info("Listing schemas for instance '%(id)s'\n"
                 "req : '%(req)s'\n\n",
                 {"id": instance_id, "req": req})

        context = req.environ[wsgi.CONTEXT_KEY]
        self.authorize_target_action(
            context, 'database:index', instance_id)
        schemas, next_marker = models.Schemas.load(context, instance_id)
        view = views.SchemasView(schemas)
        paged = pagination.SimplePaginatedDataView(req.url, 'databases', view,
                                                   next_marker)
        return wsgi.Result(paged.data(), 200)

    def create(self, req, body, tenant_id, instance_id):
        """Creates a set of schemas."""
        LOG.info("Creating schema for instance '%(id)s'\n"
                 "req : '%(req)s'\n\n"
                 "body: '%(body)s'\n'n",
                 {"id": instance_id,
                  "req": req,
                  "body": body})

        context = req.environ[wsgi.CONTEXT_KEY]
        self.authorize_target_action(
            context, 'database:create', instance_id)
        schemas = body['databases']
        context.notification = notification.DBaaSDatabaseCreate(context,
                                                                request=req)
        with StartNotification(context, instance_id=instance_id,
                               dbname=".".join([db['name']
                                                for db in schemas])):
            try:
                model_schemas = populate_validated_databases(schemas)
                models.Schema.create(context, instance_id, model_schemas)
            except (ValueError, AttributeError) as e:
                raise exception.BadRequest(_("Database create error: %(e)s")
                                           % {'e': e})
        return wsgi.Result(None, 202)

    def delete(self, req, tenant_id, instance_id, id):
        LOG.info("Deleting schema for instance '%(id)s'\n"
                 "req : '%(req)s'\n\n",
                 {"id": instance_id, "req": req})
        context = req.environ[wsgi.CONTEXT_KEY]
        self.authorize_target_action(
            context, 'database:delete', instance_id)
        context.notification = notification.DBaaSDatabaseDelete(
            context, request=req)
        with StartNotification(context, instance_id=instance_id, dbname=id):
            try:
                schema = guest_models.MySQLSchema(name=id)
                schema.check_delete()
                if not models.Schemas.find(context, instance_id, id):
                    raise exception.DatabaseNotFound(uuid=id)
                models.Schema.delete(context, instance_id, schema.serialize())
            except (ValueError, AttributeError) as e:
                raise exception.BadRequest(_("Database delete error: %(e)s")
                                           % {'e': e})
        return wsgi.Result(None, 202)

    def show(self, req, tenant_id, instance_id, id):
        context = req.environ[wsgi.CONTEXT_KEY]
        self.authorize_target_action(
            context, 'database:show', instance_id)
        raise webob.exc.HTTPNotImplemented()



# Copyright 2013 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_concurrency import processutils
from oslo_service import service as openstack_service

from trove.cmd.common import with_initialize
from trove.conductor import api as conductor_api


@with_initialize
def main(conf):
    from trove.common import notification
    from trove.common.rpc import conductor_host_serializer as sz
    from trove.common.rpc import service as rpc_service
    from trove.instance import models as inst_models

    notification.DBaaSAPINotification.register_notify_callback(
        inst_models.persist_instance_fault)
    topic = conf.conductor_queue
    server = rpc_service.RpcService(
        key=None, manager=conf.conductor_manager, topic=topic,
        rpc_api_version=conductor_api.API.API_LATEST_VERSION,
        secure_serializer=sz.ConductorHostSerializer)
    workers = conf.trove_conductor_workers or processutils.get_worker_count()
    launcher = openstack_service.launch(conf, server, workers=workers,
                                        restart_method='mutate')
    launcher.wait()



# Copyright 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import abc


class NetworkDriver(object, metaclass=abc.ABCMeta):
    """Base Network Driver class to abstract the network driver used."""

    @abc.abstractmethod
    def get_sec_group_by_id(self, group_id):
        """
        Returns security group with given group_id
        """

    @abc.abstractmethod
    def delete_security_group(self, sec_group_id):
        """Deletes the security group by given ID."""

    @abc.abstractmethod
    def delete_security_group_rule(self, sec_group_rule_id):
        """Deletes the rule by given ID."""



# Copyright [2015] Hewlett-Packard Development Company, L.P.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime

from unittest.mock import Mock
from unittest.mock import patch

from trove import rpc
from trove.cluster.models import ClusterTasks as ClusterTaskStatus
from trove.cluster.models import DBCluster
import trove.common.context as context
from trove.common.exception import GuestError
from trove.common.strategies.cluster.experimental.vertica.taskmanager import (
    VerticaClusterTasks as ClusterTasks)
from trove.common.strategies.cluster.experimental.vertica.taskmanager import (
    VerticaTaskManagerAPI as task_api)
from trove.common.strategies.cluster.experimental.vertica.taskmanager import (
    VerticaTaskManagerStrategy as task_strategy)
from trove.datastore import models as datastore_models
from trove.instance.models import BaseInstance
from trove.instance.models import DBInstance
from trove.instance.models import Instance
from trove.instance.models import InstanceServiceStatus
from trove.instance.models import InstanceTasks
from trove.instance.service_status import ServiceStatuses
from trove.tests.unittests import trove_testtools


class VerticaClusterTasksTest(trove_testtools.TestCase):
    def setUp(self):
        super(VerticaClusterTasksTest, self).setUp()
        self.cluster_id = "1232"
        self.cluster_name = "Cluster-1234"
        self.tenant_id = "6789"
        self.db_cluster = DBCluster(ClusterTaskStatus.NONE,
                                    id=self.cluster_id,
                                    created=str(datetime.date),
                                    updated=str(datetime.date),
                                    name=self.cluster_name,
                                    task_id=ClusterTaskStatus.NONE._code,
                                    tenant_id=self.tenant_id,
                                    datastore_version_id="1",
                                    deleted=False)
        self.dbinst1 = DBInstance(InstanceTasks.NONE, id="1", name="member1",
                                  compute_instance_id="compute-1",
                                  task_id=InstanceTasks.NONE._code,
                                  task_description=InstanceTasks.NONE._db_text,
                                  volume_id="volume-1",
                                  datastore_version_id="1",
                                  cluster_id=self.cluster_id,
                                  type="master")
        self.dbinst2 = DBInstance(InstanceTasks.NONE, id="2", name="member2",
                                  compute_instance_id="compute-2",
                                  task_id=InstanceTasks.NONE._code,
                                  task_description=InstanceTasks.NONE._db_text,
                                  volume_id="volume-2",
                                  datastore_version_id="1",
                                  cluster_id=self.cluster_id,
                                  type="member")
        self.dbinst3 = DBInstance(InstanceTasks.NONE, id="3", name="member3",
                                  compute_instance_id="compute-3",
                                  task_id=InstanceTasks.NONE._code,
                                  task_description=InstanceTasks.NONE._db_text,
                                  volume_id="volume-3",
                                  datastore_version_id="1",
                                  cluster_id=self.cluster_id,
                                  type="member")
        mock_ds1 = Mock()
        mock_ds1.name = 'vertica'
        mock_dv1 = Mock()
        mock_dv1.name = '7.1'
        self.clustertasks = ClusterTasks(Mock(),
                                         self.db_cluster,
                                         datastore=mock_ds1,
                                         datastore_version=mock_dv1)

    @patch.object(ClusterTasks, 'update_statuses_on_failure')
    @patch.object(DBInstance, 'find_by')
    @patch.object(InstanceServiceStatus, 'find_by')
    @patch('trove.taskmanager.models.LOG')
    def test_all_instances_ready_with_server_error(self,
                                                   mock_logging, mock_find,
                                                   mock_db_find, mock_update):
        (mock_find.return_value.
         get_status.return_value) = ServiceStatuses.NEW
        (mock_db_find.return_value.
         get_task_status.return_value) = InstanceTasks.BUILDING_ERROR_SERVER
        ret_val = self.clustertasks._all_instances_ready(["1", "2", "3", "4"],
                                                         self.cluster_id)
        mock_update.assert_called_with(self.cluster_id, None)
        self.assertFalse(ret_val)

    @patch.object(ClusterTasks, 'update_statuses_on_failure')
    @patch.object(DBInstance, 'find_by')
    @patch.object(InstanceServiceStatus, 'find_by')
    @patch('trove.taskmanager.models.LOG')
    def test_all_instances_ready_bad_status(self, mock_logging,
                                            mock_find, mock_db_find,
                                            mock_update):
        (mock_find.return_value.
         get_status.return_value) = ServiceStatuses.FAILED
        (mock_db_find.return_value.
         get_task_status.return_value) = InstanceTasks.NONE
        ret_val = self.clustertasks._all_instances_ready(["1", "2", "3", "4"],
                                                         self.cluster_id)
        mock_update.assert_called_with(self.cluster_id, None)
        self.assertFalse(ret_val)

    @patch.object(DBInstance, 'find_by')
    @patch.object(InstanceServiceStatus, 'find_by')
    def test_all_instances_ready(self, mock_find, mock_db_find):
        (mock_find.return_value.
         get_status.return_value) = ServiceStatuses.INSTANCE_READY
        (mock_db_find.return_value.
         get_task_status.return_value) = InstanceTasks.NONE
        ret_val = self.clustertasks._all_instances_ready(["1", "2", "3", "4"],
                                                         self.cluster_id)
        self.assertTrue(ret_val)

    @patch.object(ClusterTasks, 'reset_task')
    @patch.object(ClusterTasks, '_all_instances_ready', return_value=False)
    @patch.object(Instance, 'load')
    @patch.object(DBInstance, 'find_all')
    @patch.object(datastore_models.Datastore, 'load')
    @patch.object(datastore_models.DatastoreVersion, 'load_by_uuid')
    def test_create_cluster_instance_not_ready(self, mock_dv, mock_ds,
                                               mock_find_all, mock_load,
                                               mock_ready, mock_reset_task):
        mock_find_all.return_value.all.return_value = [self.dbinst1]
        mock_load.return_value = BaseInstance(Mock(),
                                              self.dbinst1, Mock(),
                                              InstanceServiceStatus(
                                                  ServiceStatuses.NEW))
        self.clustertasks.create_cluster(Mock(), self.cluster_id)
        mock_reset_task.assert_called_with()

    @patch.object(ClusterTasks, 'reset_task')
    @patch.object(ClusterTasks, 'get_guest')
    @patch.object(ClusterTasks, 'get_ip')
    @patch.object(ClusterTasks, '_all_instances_ready')
    @patch.object(Instance, 'load')
    @patch.object(DBInstance, 'find_all')
    @patch.object(datastore_models.Datastore, 'load')
    @patch.object(datastore_models.DatastoreVersion, 'load_by_uuid')
    def test_create_cluster(self, mock_dv, mock_ds, mock_find_all, mock_load,
                            mock_ready, mock_ip, mock_guest, mock_reset_task):
        cluster_instances = [self.dbinst1, self.dbinst2, self.dbinst3]
        for instance in cluster_instances:
            if instance['type'] == "master":
                mock_find_all.return_value.all.return_value = [self.dbinst1]
            mock_ready.return_value = True
        mock_load.return_value = BaseInstance(Mock(),
                                              self.dbinst1, Mock(),
                                              InstanceServiceStatus(
                                                  ServiceStatuses.NEW))
        mock_ip.return_value = "10.0.0.2"
        self.clustertasks.create_cluster(Mock(), self.cluster_id)
        mock_guest.return_value.install_cluster.assert_called_with(
            ['10.0.0.2'])
        mock_reset_task.assert_called_with()
        mock_guest.return_value.cluster_complete.assert_called_with()

    @patch.object(ClusterTasks, 'update_statuses_on_failure')
    @patch.object(ClusterTasks, 'reset_task')
    @patch.object(ClusterTasks, 'get_ip')
    @patch.object(ClusterTasks, '_all_instances_ready')
    @patch.object(Instance, 'load')
    @patch.object(DBInstance, 'find_all')
    @patch.object(datastore_models.Datastore, 'load')
    @patch.object(datastore_models.DatastoreVersion, 'load_by_uuid')
    @patch(
        'trove.common.strategies.cluster.experimental.vertica.taskmanager.LOG')
    def test_create_cluster_fail(self, mock_logging, mock_dv, mock_ds,
                                 mock_find_all,
                                 mock_load, mock_ready, mock_ip,
                                 mock_reset_task, mock_update_status):
        mock_find_all.return_value.all.return_value = [self.dbinst1]
        mock_load.return_value = BaseInstance(Mock(),
                                              self.dbinst1, Mock(),
                                              InstanceServiceStatus(
                                                  ServiceStatuses.NEW))
        mock_ip.return_value = "10.0.0.2"
        guest_client = Mock()
        guest_client.install_cluster = Mock(side_effect=GuestError("Error"))
        with patch.object(ClusterTasks, 'get_guest',
                          return_value=guest_client):
            self.clustertasks.create_cluster(Mock(), self.cluster_id)
            mock_update_status.assert_called_with('1232')
            mock_reset_task.assert_called_with()


class VerticaTaskManagerAPITest(trove_testtools.TestCase):
    @patch.object(rpc, 'get_client', Mock(return_value=Mock()))
    def setUp(self):
        super(VerticaTaskManagerAPITest, self).setUp()
        self.context = context.TroveContext()
        self.api = task_api(self.context)
        self.call_context = trove_testtools.TroveTestContext(self)
        self.api.client.prepare = Mock(return_value=self.call_context)
        self.call_context.cast = Mock()
        self.rpc_api_version = '1.0'

    def test_task_manager_api_cast(self):
        self.api._cast(method_name='test_method', version=self.rpc_api_version)
        self.call_context.cast.assert_called_with(self.context, 'test_method')


class VerticaTaskManagerStrategyTest(trove_testtools.TestCase):

    def test_task_manager_cluster_tasks_class(self):
        vertica_strategy = task_strategy()
        self.assertFalse(
            hasattr(vertica_strategy.task_manager_cluster_tasks_class,
                    'rebuild_cluster'))
        self.assertTrue(callable(
            vertica_strategy.task_manager_cluster_tasks_class.create_cluster))

    def test_task_manager_api_class(self):
        vertica_strategy = task_strategy()
        self.assertFalse(hasattr(vertica_strategy.task_manager_api_class,
                                 'add_new_node'))
        self.assertTrue(
            callable(vertica_strategy.task_manager_api_class._cast))



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# Encryption/decryption handling

import hashlib
import os
from oslo_utils import encodeutils
import random
import string

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers import modes
from trove.common import stream_codecs


IV_BYTE_COUNT = 16
_CRYPT_BACKEND = None


def _get_cipher(key, iv):
    global _CRYPT_BACKEND
    if not _CRYPT_BACKEND:
        _CRYPT_BACKEND = default_backend()

    return Cipher(algorithms.AES(key), modes.CBC(iv),
                  backend=_CRYPT_BACKEND)


def _encrypt(key, iv, data):
    encryptor = _get_cipher(key, iv).encryptor()
    return encryptor.update(data) + encryptor.finalize()


def _decrypt(key, iv, data):
    decryptor = _get_cipher(key, iv).decryptor()
    return decryptor.update(data) + decryptor.finalize()


def encode_data(data):
    # NOTE(zhaochao) No need to encoding string object any more,
    # as Base64Codec is now using oslo_serialization.base64 which
    # could take care of this.
    return stream_codecs.Base64Codec().serialize(data)


def decode_data(data):
    return stream_codecs.Base64Codec().deserialize(data)


# Pad the data string to an multiple of pad_size
def pad_for_encryption(data, pad_size=IV_BYTE_COUNT):
    pad_count = pad_size - (len(data) % pad_size)
    return data + bytes((pad_count,)) * pad_count


# Unpad the data string by stripping off excess characters
def unpad_after_decryption(data):
    return data[:len(data) - data[-1]]


def encrypt_data(data, key, iv_byte_count=IV_BYTE_COUNT):
    data = encodeutils.to_utf8(data)
    key = encodeutils.to_utf8(key)
    md5_key = encodeutils.safe_encode(hashlib.md5(key).hexdigest())
    iv = os.urandom(iv_byte_count)
    iv = iv[:iv_byte_count]
    data = pad_for_encryption(data, iv_byte_count)
    encrypted = _encrypt(md5_key, bytes(iv), data)
    return iv + encrypted


def decrypt_data(data, key, iv_byte_count=IV_BYTE_COUNT):
    key = encodeutils.to_utf8(key)
    md5_key = encodeutils.safe_encode(hashlib.md5(key).hexdigest())
    iv = data[:iv_byte_count]
    decrypted = _decrypt(md5_key, bytes(iv), bytes(data[iv_byte_count:]))
    return unpad_after_decryption(decrypted)


def generate_random_key(length=32, chars=None):
    chars = chars if chars else (string.ascii_uppercase +
                                 string.ascii_lowercase + string.digits)
    return ''.join(random.choice(chars) for _ in range(length))



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from proboscis import test

from trove.tests.scenario import groups
from trove.tests.scenario.groups.test_group import TestGroup
from trove.tests.scenario.runners import test_runners


GROUP = "scenario.replication_group"


class ReplicationRunnerFactory(test_runners.RunnerFactory):

    _runner_ns = 'replication_runners'
    _runner_cls = 'ReplicationRunner'


class BackupRunnerFactory(test_runners.RunnerFactory):

    _runner_ns = 'backup_runners'
    _runner_cls = 'BackupRunner'


@test(depends_on_groups=[groups.INST_CREATE],
      groups=[GROUP, groups.REPL_INST_CREATE])
class ReplicationInstCreateGroup(TestGroup):
    """Test Replication Instance Create functionality."""

    def __init__(self):
        super(ReplicationInstCreateGroup, self).__init__(
            ReplicationRunnerFactory.instance())

    @test
    def add_data_for_replication(self):
        """Add data to master for initial replica setup."""
        self.test_runner.run_add_data_for_replication()

    @test(depends_on=[add_data_for_replication])
    def verify_data_for_replication(self):
        """Verify initial data exists on master."""
        self.test_runner.run_verify_data_for_replication()

    @test(runs_after=[verify_data_for_replication])
    def create_non_affinity_master(self):
        """Test creating a non-affinity master."""
        self.test_runner.run_create_non_affinity_master()

    @test(runs_after=[create_non_affinity_master])
    def create_single_replica(self):
        """Test creating a single replica."""
        self.test_runner.run_create_single_replica()


@test(depends_on_classes=[ReplicationInstCreateGroup],
      groups=[GROUP, groups.REPL_INST_CREATE_WAIT])
class ReplicationInstCreateWaitGroup(TestGroup):
    """Wait for Replication Instance Create to complete."""

    def __init__(self):
        super(ReplicationInstCreateWaitGroup, self).__init__(
            ReplicationRunnerFactory.instance())

    @test
    def wait_for_non_affinity_master(self):
        """Wait for non-affinity master to complete."""
        self.test_runner.run_wait_for_non_affinity_master()

    @test(depends_on=[wait_for_non_affinity_master])
    def create_non_affinity_replica(self):
        """Test creating a non-affinity replica."""
        self.test_runner.run_create_non_affinity_replica()

    @test(depends_on=[create_non_affinity_replica])
    def wait_for_non_affinity_replica_fail(self):
        """Wait for non-affinity replica to fail."""
        self.test_runner.run_wait_for_non_affinity_replica_fail()

    @test(runs_after=[wait_for_non_affinity_replica_fail])
    def delete_non_affinity_repl(self):
        """Test deleting non-affinity replica."""
        self.test_runner.run_delete_non_affinity_repl()

    @test(runs_after=[delete_non_affinity_repl])
    def wait_for_single_replica(self):
        """Wait for single replica to complete."""
        self.test_runner.run_wait_for_single_replica()

    @test(depends_on=[wait_for_single_replica])
    def add_data_after_replica(self):
        """Add data to master after initial replica is setup"""
        self.test_runner.run_add_data_after_replica()

    @test(depends_on=[add_data_after_replica])
    def verify_replica_data_after_single(self):
        """Verify data exists on single replica"""
        self.test_runner.run_verify_replica_data_after_single()


@test(depends_on_classes=[ReplicationInstCreateWaitGroup],
      groups=[GROUP, groups.REPL_INST_MULTI_CREATE])
class ReplicationInstMultiCreateGroup(TestGroup):
    """Test Replication Instance Multi-Create functionality."""

    def __init__(self):
        super(ReplicationInstMultiCreateGroup, self).__init__(
            ReplicationRunnerFactory.instance())
        self.backup_runner = BackupRunnerFactory.instance()

    @test
    def backup_master_instance(self):
        """Backup the master instance."""
        self.backup_runner.run_backup_create()
        self.backup_runner.run_backup_create_completed()
        self.test_runner.master_backup_count += 1

    @test(depends_on=[backup_master_instance])
    def create_multiple_replicas(self):
        """Test creating multiple replicas."""
        self.test_runner.run_create_multiple_replicas()

    @test(depends_on=[create_multiple_replicas])
    def check_has_incremental_backup(self):
        """Test that creating multiple replicas uses incr backup."""
        self.backup_runner.run_check_has_incremental()


@test(depends_on_classes=[ReplicationInstMultiCreateGroup],
      groups=[GROUP, groups.REPL_INST_DELETE_NON_AFFINITY_WAIT])
class ReplicationInstDeleteNonAffReplWaitGroup(TestGroup):
    """Wait for Replication Instance Non-Affinity repl to be gone."""

    def __init__(self):
        super(ReplicationInstDeleteNonAffReplWaitGroup, self).__init__(
            ReplicationRunnerFactory.instance())

    @test
    def wait_for_delete_non_affinity_repl(self):
        """Wait for the non-affinity replica to delete."""
        self.test_runner.run_wait_for_delete_non_affinity_repl()

    @test(depends_on=[wait_for_delete_non_affinity_repl])
    def delete_non_affinity_master(self):
        """Test deleting non-affinity master."""
        self.test_runner.run_delete_non_affinity_master()


@test(depends_on_classes=[ReplicationInstDeleteNonAffReplWaitGroup],
      groups=[GROUP, groups.REPL_INST_MULTI_CREATE_WAIT])
class ReplicationInstMultiCreateWaitGroup(TestGroup):
    """Wait for Replication Instance Multi-Create to complete."""

    def __init__(self):
        super(ReplicationInstMultiCreateWaitGroup, self).__init__(
            ReplicationRunnerFactory.instance())

    @test
    def wait_for_delete_non_affinity_master(self):
        """Wait for the non-affinity master to delete."""
        self.test_runner.run_wait_for_delete_non_affinity_master()

    @test(runs_after=[wait_for_delete_non_affinity_master])
    def wait_for_multiple_replicas(self):
        """Wait for multiple replicas to complete."""
        self.test_runner.run_wait_for_multiple_replicas()

    @test(depends_on=[wait_for_multiple_replicas])
    def verify_replica_data_orig(self):
        """Verify original data was transferred to replicas."""
        self.test_runner.run_verify_replica_data_orig()

    @test(depends_on=[wait_for_multiple_replicas],
          runs_after=[verify_replica_data_orig])
    def add_data_to_replicate(self):
        """Add new data to master to verify replication."""
        self.test_runner.run_add_data_to_replicate()

    @test(depends_on=[add_data_to_replicate])
    def verify_data_to_replicate(self):
        """Verify new data exists on master."""
        self.test_runner.run_verify_data_to_replicate()

    @test(depends_on=[add_data_to_replicate],
          runs_after=[verify_data_to_replicate])
    def verify_replica_data_orig2(self):
        """Verify original data was transferred to replicas."""
        self.test_runner.run_verify_replica_data_orig()

    @test(depends_on=[add_data_to_replicate],
          runs_after=[verify_replica_data_orig2])
    def verify_replica_data_new(self):
        """Verify new data was transferred to replicas."""
        self.test_runner.run_verify_replica_data_new()

    @test(depends_on=[wait_for_multiple_replicas],
          runs_after=[verify_replica_data_new])
    def promote_master(self):
        """Ensure promoting master fails."""
        self.test_runner.run_promote_master()

    @test(depends_on=[wait_for_multiple_replicas],
          runs_after=[promote_master])
    def eject_replica(self):
        """Ensure ejecting non master fails."""
        self.test_runner.run_eject_replica()

    @test(depends_on=[wait_for_multiple_replicas],
          runs_after=[eject_replica])
    def eject_valid_master(self):
        """Ensure ejecting valid master fails."""
        self.test_runner.run_eject_valid_master()

    @test(depends_on=[wait_for_multiple_replicas],
          runs_after=[eject_valid_master])
    def delete_valid_master(self):
        """Ensure deleting valid master fails."""
        self.test_runner.run_delete_valid_master()


@test(depends_on_classes=[ReplicationInstMultiCreateWaitGroup],
      groups=[GROUP, groups.REPL_INST_MULTI_PROMOTE])
class ReplicationInstMultiPromoteGroup(TestGroup):
    """Test Replication Instance Multi-Promote functionality."""

    def __init__(self):
        super(ReplicationInstMultiPromoteGroup, self).__init__(
            ReplicationRunnerFactory.instance())

    @test
    def promote_to_replica_source(self):
        """Test promoting a replica to replica source (master)."""
        self.test_runner.run_promote_to_replica_source()

    @test(depends_on=[promote_to_replica_source])
    def verify_replica_data_new_master(self):
        """Verify data is still on new master."""
        self.test_runner.run_verify_replica_data_new_master()

    @test(depends_on=[promote_to_replica_source],
          runs_after=[verify_replica_data_new_master])
    def add_data_to_replicate2(self):
        """Add data to new master to verify replication."""
        self.test_runner.run_add_data_to_replicate2()

    @test(depends_on=[add_data_to_replicate2])
    def verify_data_to_replicate2(self):
        """Verify data exists on new master."""
        self.test_runner.run_verify_data_to_replicate2()

    @test(depends_on=[add_data_to_replicate2],
          runs_after=[verify_data_to_replicate2])
    def verify_replica_data_new2(self):
        """Verify data was transferred to new replicas."""
        self.test_runner.run_verify_replica_data_new2()

    @test(depends_on=[promote_to_replica_source],
          runs_after=[verify_replica_data_new2])
    def promote_original_source(self):
        """Test promoting back the original replica source."""
        self.test_runner.run_promote_original_source()

    @test(depends_on=[promote_original_source])
    def add_final_data_to_replicate(self):
        """Add final data to original master to verify switch."""
        self.test_runner.run_add_final_data_to_replicate()

    @test(depends_on=[add_final_data_to_replicate])
    def verify_data_to_replicate_final(self):
        """Verify final data exists on master."""
        self.test_runner.run_verify_data_to_replicate_final()

    @test(depends_on=[verify_data_to_replicate_final])
    def verify_final_data_replicated(self):
        """Verify final data was transferred to all replicas."""
        self.test_runner.run_verify_final_data_replicated()


@test(depends_on_classes=[ReplicationInstMultiPromoteGroup],
      groups=[GROUP, groups.REPL_INST_DELETE])
class ReplicationInstDeleteGroup(TestGroup):
    """Test Replication Instance Delete functionality."""

    def __init__(self):
        super(ReplicationInstDeleteGroup, self).__init__(
            ReplicationRunnerFactory.instance())

    @test
    def remove_replicated_data(self):
        """Remove replication data."""
        self.test_runner.run_remove_replicated_data()

    @test(runs_after=[remove_replicated_data])
    def detach_replica_from_source(self):
        """Test detaching a replica from the master."""
        self.test_runner.run_detach_replica_from_source()

    @test(runs_after=[detach_replica_from_source])
    def delete_detached_replica(self):
        """Test deleting the detached replica."""
        self.test_runner.run_delete_detached_replica()

    @test(runs_after=[delete_detached_replica])
    def delete_all_replicas(self):
        """Test deleting all the remaining replicas."""
        self.test_runner.run_delete_all_replicas()


@test(depends_on_classes=[ReplicationInstDeleteGroup],
      groups=[GROUP, groups.REPL_INST_DELETE_WAIT])
class ReplicationInstDeleteWaitGroup(TestGroup):
    """Wait for Replication Instance Delete to complete."""

    def __init__(self):
        super(ReplicationInstDeleteWaitGroup, self).__init__(
            ReplicationRunnerFactory.instance())
        self.backup_runner = BackupRunnerFactory.instance()

    @test
    def wait_for_delete_replicas(self):
        """Wait for all the replicas to delete."""
        self.test_runner.run_wait_for_delete_replicas()

    @test(runs_after=[wait_for_delete_replicas])
    def test_backup_deleted(self):
        """Remove the full backup and test that the created backup
           is now gone.
        """
        self.test_runner.run_test_backup_deleted()
        self.backup_runner.run_delete_backup()

    @test(runs_after=[test_backup_deleted])
    def cleanup_master_instance(self):
        """Remove slave users from master instance."""
        self.test_runner.run_cleanup_master_instance()



# Copyright 2014 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from oslo_log import log as logging

from trove.guestagent.strategies.replication import mysql_base

LOG = logging.getLogger(__name__)


class MysqlGTIDReplication(mysql_base.MysqlReplicationBase):
    """MySql Replication coordinated by GTIDs."""
    def connect_to_master(self, service, master_info):
        if 'dataset' in master_info:
            # pull the last executed GTID from the master via
            # the xtrabackup metadata file. If that value is
            # provided we need to set the gtid_purged variable
            # before executing the CHANGE MASTER TO command
            last_gtid = self.read_last_master_gtid(service)
            LOG.info("last_gtid value is %s", last_gtid)
            if '-' in last_gtid:
                # See
                # https://avdeo.com/tag/error-1840-hy000-global-gtid_purged-can-only-be-set-when/
                # Also, FLUSH PRIVILEGES will restore gtid_executed.
                service.execute_sql('RESET MASTER')
                set_gtid_cmd = "SET GLOBAL gtid_purged='%s'" % last_gtid
                service.execute_sql(set_gtid_cmd)

        replica_conf = master_info['replica_conf']
        LOG.info(
            "Configure the slave, master: %s:%s, replication user: %s",
            master_info['master']['host'],
            master_info['master']['port'],
            replica_conf['replication_user']['name']
        )

        change_master_cmd = (
            "CHANGE MASTER TO "
            "MASTER_HOST='%(host)s', "
            "MASTER_PORT=%(port)s, "
            "MASTER_USER='%(user)s', "
            "MASTER_PASSWORD='%(password)s', "
            "MASTER_AUTO_POSITION=1, "
            "MASTER_CONNECT_RETRY=15" %
            {
                'host': master_info['master']['host'],
                'port': master_info['master']['port'],
                'user': replica_conf['replication_user']['name'],
                'password': replica_conf['replication_user']['password']
            })
        service.execute_sql(change_master_cmd)

        service.start_slave()



# Copyright 2021 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""The code related to integration between oslo.cache module and trove."""

from oslo_cache import core
from oslo_config import cfg


def register_cache_configurations(conf):
    """Register all configurations required for oslo.cache.

    The procedure registers all configurations required for oslo.cache.
    It should be called before configuring of cache region
    """
    core.configure(conf)

    ports_cache_group = cfg.OptGroup('instance_ports_cache')
    ports_cache_opts = [
        cfg.IntOpt('expiration_time', default=86400,
                   help='TTL, in seconds, for any cached item in the '
                        'dogpile.cache region used for caching of the '
                        'instance ports.'),
        cfg.BoolOpt("caching", default=True,
                    help='Toggle to enable/disable caching when getting trove '
                         'instance ports. Please note that the global toggle '
                         'for oslo.cache(enabled=True in [cache] group) '
                         'must be enabled to use this feature.')
    ]
    conf.register_group(ports_cache_group)
    conf.register_opts(ports_cache_opts, group=ports_cache_group)

    return conf


# variable that stores an initialized cache region for trove
_REGION = None


def get_cache_region():
    global _REGION
    if not _REGION:
        _REGION = core.configure_cache_region(
            conf=register_cache_configurations(cfg.CONF),
            region=core.create_region())
    return _REGION



# Copyright (c) 2011 OpenStack, LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Functions to initiate and shut down services needed by the tests."""

import os
import re
import subprocess
import time

from collections import namedtuple
from httplib2 import Http
from nose.plugins.skip import SkipTest

from proboscis import decorators


def _is_web_service_alive(url):
    """Does a HTTP GET request to see if the web service is up."""
    client = Http()
    try:
        resp = client.request(url, 'GET')
        return resp != None
    except Exception:
        return False


_running_services = []


def get_running_services():
    """ Returns the list of services which this program has started."""
    return _running_services


def start_proc(cmd, shell=False):
    """Given a command, starts and returns a process."""
    env = os.environ.copy()
    proc = subprocess.Popen(
        cmd,
        shell=shell,
        stdin=subprocess.PIPE,
        bufsize=0,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=env
    )
    return proc


MemoryInfo = namedtuple("MemoryInfo", ['mapped', 'writeable', 'shared'])


class Service(object):
    """Starts and stops a service under test.

    The methods to start and stop the service will not actually do anything
    if they detect the service is already running on this machine.  This is
    because it may be useful for developers to start the services themselves
    some other way.

    """

    # TODO(tim.simpson): Hard to follow, consider renaming certain attributes.

    def __init__(self, cmd):
        """Defines a service to run."""
        if not isinstance(cmd, list):
            raise TypeError()
        self.cmd = cmd
        self.do_not_manage_proc = False
        self.proc = None

    def __del__(self):
        if self.is_running:
            self.stop()

    def ensure_started(self):
        """Starts the service if it is not running."""
        if not self.is_running:
            self.start()

    def find_proc_id(self):
        """Finds and returns the process id."""
        if not self.cmd:
            return False
        # The cmd[1] signifies the executable python script. It gets invoked
        # as python /path/to/executable args, so the entry is
        # /path/to/executable
        actual_command = self.cmd[1].split("/")[-1]
        proc_command = ["/usr/bin/pgrep", "-f", actual_command]
        proc = start_proc(proc_command, shell=False)
        # this is to make sure there is only one pid returned from the pgrep
        has_two_lines = False
        pid = None
        for line in iter(proc.stdout.readline, ""):
            if has_two_lines:
                raise RuntimeError("Found PID twice.")
            pid = int(line)
            has_two_lines = True
        return pid

    def get_memory_info(self):
        """Returns how much memory the process is using according to pmap."""
        pid = self.find_proc_id()
        if not pid:
            raise RuntimeError("Can't find PID, so can't get memory.")
        proc = start_proc(["/usr/bin/pmap", "-d", str(pid)],
                          shell=False)
        for line in iter(proc.stdout.readline, ""):
            m = re.search(r"mapped\:\s([0-9]+)K\s+"
                          r"writeable/private:\s([0-9]+)K\s+"
                          r"shared:\s+([0-9]+)K", line)
            if m:
                return MemoryInfo(int(m.group(1)), int(m.group(2)),
                                  int(m.group(3)))
        raise RuntimeError("Memory info not found.")

    def get_fd_count_from_proc_file(self):
        """Returns file descriptors according to /proc/<id>/status."""
        pid = self.find_proc_id()
        with open("/proc/%d/status" % pid) as status:
            for line in status.readlines():
                index = line.find(":")
                name = line[:index]
                value = line[index + 1:]
                if name == "FDSize":
                    return int(value)
        raise RuntimeError("FDSize not found!")

    def get_fd_count(self):
        """Returns file descriptors according to /proc/<id>/status."""
        pid = self.find_proc_id()
        cmd = "Finding file descriptors..."
        print("CMD" + cmd)
        proc = start_proc(['ls', '-la', '/proc/%d/fd' % pid], shell=False)
        count = -3
        has_two_lines = False
        for line in iter(proc.stdout.readline, ""):
            print("\t" + line)
            count += 1
        if not count:
            raise RuntimeError("Could not get file descriptors!")
        return count


        with open("/proc/%d/fd" % pid) as status:
            for line in status.readlines():
                index = line.find(":")
                name = line[:index]
                value = line[index + 1:]
                if name == "FDSize":
                    return int(value)
        raise RuntimeError("FDSize not found!")

    def kill_proc(self):
        """Kills the process, wherever it may be."""
        pid = self.find_proc_id()
        if pid:
            start_proc("sudo kill -9 " + str(pid), shell=True)
            time.sleep(1)
        if self.is_service_alive():
            raise RuntimeError('Cannot kill process, PID=' +
                               str(self.proc.pid))

    def is_service_alive(self, proc_name_index=1):
        """Searches for the process to see if its alive.

         This function will return true even if this class has not started
         the service (searches using ps).

         """
        if not self.cmd:
            return False
        time.sleep(1)
        # The cmd[1] signifies the executable python script. It gets invoked
        # as python /path/to/executable args, so the entry is
        # /path/to/executable
        actual_command = self.cmd[proc_name_index].split("/")[-1]
        print(actual_command)
        proc_command = ["/usr/bin/pgrep", "-f", actual_command]
        print(proc_command)
        proc = start_proc(proc_command, shell=False)
        line = proc.stdout.readline()
        print(line)
        # pgrep only returns a pid. if there is no pid, it'll return nothing
        return len(line) != 0

    @property
    def is_running(self):
        """Returns true if the service has already been started.

        Returns true if this program has started the service or if it
        previously detected it had started.  The main use of this property
        is to know if the service was already begun by this program-
        use is_service_alive for a more definitive answer.

        """
        return self.proc or self.do_not_manage_proc

    def restart(self, extra_args):
        if self.do_not_manage_proc:
            raise RuntimeError("Can't restart proc as the tests don't own it.")
        self.stop()
        time.sleep(2)
        self.start(extra_args=extra_args)

    def start(self, time_out=30, extra_args=None):
        """Starts the service if necessary."""
        extra_args = extra_args or []
        if self.is_running:
            raise RuntimeError("Process is already running.")
        if self.is_service_alive():
            self.do_not_manage_proc = True
            return
        self.proc = start_proc(self.cmd + extra_args, shell=False)
        if not self._wait_for_start(time_out=time_out):
            self.stop()
            raise RuntimeError("Issued the command successfully but the "
                               "service (" + str(self.cmd + extra_args) +
                               ") never seemed to start.")
        _running_services.append(self)

    def stop(self):
        """Stops the service, but only if this program started it."""
        if self.do_not_manage_proc:
            return
        if not self.proc:
            raise RuntimeError("Process was not started.")
        self.proc.terminate()
        self.proc.kill()
        self.proc.wait()
        self.proc.stdin.close()
        self.kill_proc()
        self.proc = None
        global _running_services
        _running_services = [svc for svc in _running_services if svc != self]

    def _wait_for_start(self, time_out):
        """Waits until time_out (in seconds) for service to appear."""
        give_up_time = time.time() + time_out
        while time.time() < give_up_time:
            if self.is_service_alive():
                return True
        return False


class NativeService(Service):

    def is_service_alive(self):
        return super(NativeService, self).is_service_alive(proc_name_index=0)



class WebService(Service):
    """Starts and stops a web service under test."""

    def __init__(self, cmd, url):
        """Defines a service to run."""
        Service.__init__(self, cmd)
        if not isinstance(url, (str, unicode)):
            raise TypeError()
        self.url = url
        self.do_not_manage_proc = self.is_service_alive()

    def is_service_alive(self):
        """Searches for the process to see if its alive."""
        return _is_web_service_alive(self.url)



# Copyright 2013 OpenStack Foundation
# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import strutils

from trove.backup import views
from trove.backup.models import Backup
from trove.backup.models import BackupStrategy
from trove.common import apischema
from trove.common import exception
from trove.common import notification
from trove.common import pagination
from trove.common import policy
from trove.common import utils
from trove.common import wsgi
from trove.common.notification import StartNotification

LOG = logging.getLogger(__name__)


class BackupController(wsgi.Controller):
    """
    Controller for accessing backups in the OpenStack API.
    """
    schemas = apischema.backup

    def index(self, req, tenant_id):
        """
        Return all backups information for a tenant ID.
        """
        LOG.debug("Listing backups for tenant %s", tenant_id)
        datastore = req.GET.get('datastore')
        instance_id = req.GET.get('instance_id')
        project_id = req.GET.get('project_id')
        all_projects = strutils.bool_from_string(req.GET.get('all_projects'))
        context = req.environ[wsgi.CONTEXT_KEY]

        if project_id or all_projects:
            policy.authorize_on_tenant(context, 'backup:index:all_projects')
        else:
            policy.authorize_on_tenant(context, 'backup:index')

        backups, marker = Backup.list(
            context,
            datastore=datastore,
            instance_id=instance_id,
            project_id=project_id,
            all_projects=all_projects
        )
        view = views.BackupViews(backups)
        paged = pagination.SimplePaginatedDataView(req.url, 'backups', view,
                                                   marker)
        return wsgi.Result(paged.data(), 200)

    def show(self, req, tenant_id, id):
        """Return a single backup."""
        LOG.debug("Showing a backup for tenant %(tenant_id)s ID: '%(id)s'",
                  {'tenant_id': tenant_id, 'id': id})
        context = req.environ[wsgi.CONTEXT_KEY]
        backup = Backup.get_by_id(context, id)
        policy.authorize_on_target(context, 'backup:show',
                                   {'tenant': backup.tenant_id})
        return wsgi.Result(views.BackupView(backup).data(), 200)

    def create(self, req, body, tenant_id):
        LOG.info("Creating a backup for tenant %s", tenant_id)
        context = req.environ[wsgi.CONTEXT_KEY]
        policy.authorize_on_tenant(context, 'backup:create')
        data = body['backup']
        instance = data.get('instance')
        name = data['name']
        desc = data.get('description')
        parent = data.get('parent_id')
        incremental = data.get('incremental')
        swift_container = data.get('swift_container')
        restore_from = data.get('restore_from')

        if swift_container:
            utils.validate_command(swift_container)

        context.notification = notification.DBaaSBackupCreate(
            context, request=req)

        if not restore_from:
            if not instance:
                raise exception.BackupCreationError('instance is missing.')
            if not swift_container:
                instance_id = utils.get_id_from_href(instance)
                backup_strategy = BackupStrategy.get(context, instance_id)
                if backup_strategy:
                    swift_container = backup_strategy.swift_container

        with StartNotification(context, name=name, instance_id=instance,
                               description=desc, parent_id=parent):
            backup = Backup.create(context, instance, name, desc,
                                   parent_id=parent, incremental=incremental,
                                   swift_container=swift_container,
                                   restore_from=restore_from)

        return wsgi.Result(views.BackupView(backup).data(), 202)

    def delete(self, req, tenant_id, id):
        LOG.info('Deleting backup for tenant %(tenant_id)s '
                 'ID: %(backup_id)s',
                 {'tenant_id': tenant_id, 'backup_id': id})
        context = req.environ[wsgi.CONTEXT_KEY]
        backup = Backup.get_by_id(context, id)
        policy.authorize_on_target(context, 'backup:delete',
                                   {'tenant': backup.tenant_id})
        context.notification = notification.DBaaSBackupDelete(context,
                                                              request=req)
        with StartNotification(context, backup_id=id):
            Backup.delete(context, id)
        return wsgi.Result(None, 202)


class BackupStrategyController(wsgi.Controller):
    schemas = apischema.backup_strategy

    def create(self, req, body, tenant_id):
        LOG.info("Creating or updating a backup strategy for tenant %s, "
                 "body: %s", tenant_id, body)
        context = req.environ[wsgi.CONTEXT_KEY]
        policy.authorize_on_tenant(context, 'backup_strategy:create')
        data = body['backup_strategy']

        instance_id = data.get('instance_id', '')
        swift_container = data.get('swift_container')

        backup_strategy = BackupStrategy.create(context, instance_id,
                                                swift_container)
        return wsgi.Result(
            views.BackupStrategyView(backup_strategy).data(), 202)

    def index(self, req, tenant_id):
        context = req.environ[wsgi.CONTEXT_KEY]
        instance_id = req.GET.get('instance_id')
        tenant_id = req.GET.get('project_id', context.project_id)
        LOG.info("Listing backup strateies for tenant %s", tenant_id)

        if tenant_id != context.project_id and not context.is_admin:
            raise exception.TroveOperationAuthError(
                tenant_id=context.project_id
            )
        policy.authorize_on_tenant(context, 'backup_strategy:index')

        result = BackupStrategy.list(context, tenant_id,
                                     instance_id=instance_id)
        view = views.BackupStrategiesView(result)
        return wsgi.Result(view.data(), 200)

    def delete(self, req, tenant_id):
        context = req.environ[wsgi.CONTEXT_KEY]
        instance_id = req.GET.get('instance_id', '')
        tenant_id = req.GET.get('project_id', context.project_id)
        LOG.info('Deleting backup strategies for tenant %s, instance_id=%s',
                 tenant_id, instance_id)

        if tenant_id != context.project_id and not context.is_admin:
            raise exception.TroveOperationAuthError(
                tenant_id=context.project_id
            )
        policy.authorize_on_tenant(context, 'backup_strategy:delete')

        BackupStrategy.delete(context, tenant_id, instance_id)

        return wsgi.Result(None, 202)



# Copyright 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import sqlalchemy
from sqlalchemy import schema

from trove.db.sqlalchemy.migrate_repo import schema as trove_schema

meta = schema.MetaData()


def upgrade(migrate_engine):
    meta.bind = migrate_engine

    ds_config_param = trove_schema.Table('datastore_configuration_parameters',
                                         meta,
                                         autoload=True)

    # Remove records with deleted=1
    if 'deleted' in ds_config_param.c:
        ds_config_param.delete(). \
            where(ds_config_param.c.deleted == 1). \
            execute()

        # Delete columns deleted and deleted_at
        if migrate_engine.name != "sqlite":
            ds_config_param.drop_column('deleted')
            ds_config_param.drop_column('deleted_at')
        else:
            # It is not possible to remove a column from a table in SQLite.
            # SQLite is just for testing, so we re-create the table.
            ds_config_param.drop()
            meta.clear()
            trove_schema.Table('datastore_versions', meta, autoload=True)
            new_table = trove_schema.Table(
                'datastore_configuration_parameters',
                meta,
                schema.Column('id', trove_schema.String(36),
                              primary_key=True, nullable=False),
                schema.Column('name', trove_schema.String(128),
                              primary_key=True, nullable=False),
                schema.Column('datastore_version_id', trove_schema.String(36),
                              sqlalchemy.ForeignKey("datastore_versions.id"),
                              primary_key=True, nullable=False),
                schema.Column('restart_required', trove_schema.Boolean(),
                              nullable=False, default=False),
                schema.Column('max_size', trove_schema.String(40)),
                schema.Column('min_size', trove_schema.String(40)),
                schema.Column('data_type', trove_schema.String(128),
                              nullable=False),
                schema.UniqueConstraint(
                    'datastore_version_id', 'name',
                    name=('UQ_datastore_configuration_parameters_datastore_'
                          'version_id_name')
                )
            )
            trove_schema.create_tables([new_table])



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from trove.common.db import models


class CouchDBSchema(models.DatastoreSchema):
    """Represents a CouchDB schema and its associated properties."""

    @property
    def _max_schema_name_length(self):
        return 32


class CouchDBUser(models.DatastoreUser):
    """Represents a CouchDB user and its associated properties."""

    @property
    def schema_model(self):
        return CouchDBSchema



# Copyright 2010-2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Model classes that form the core of instances functionality."""

from oslo_utils.importutils import import_class

from trove.common import cfg
from trove.common import clients
from trove.common.i18n import _

CONF = cfg.CONF


class ModelBase(object):
    """
    An object which can be stored in the database.
    """

    _data_fields = []
    _auto_generated_attrs = []

    def _validate(self, errors):
        """Subclasses override this to offer additional validation.

        For each validation error a key with the field name and an error
        message is added to the dict.

        """
        pass

    def data(self, **options):
        """Called to serialize object to a dictionary."""
        data_fields = self._data_fields + self._auto_generated_attrs
        return {field: self[field] for field in data_fields}

    def is_valid(self):
        """Called when persisting data to ensure the format is correct."""
        self.errors = {}
        self._validate(self.errors)
#        self._validate_columns_type()
#        self._before_validate()
#        self._validate()
        return self.errors == {}

    def __setitem__(self, key, value):
        """Overloaded to cause this object to look like a data entity."""
        setattr(self, key, value)

    def __getitem__(self, key):
        """Overloaded to cause this object to look like a data entity."""
        return getattr(self, key)

    def __eq__(self, other):
        """Overloaded to cause this object to look like a data entity."""
        if not hasattr(other, 'id'):
            return False
        return type(other) == type(self) and other.id == self.id

    def __ne__(self, other):
        """Overloaded to cause this object to look like a data entity."""
        return not self == other

    def __hash__(self):
        """Overloaded to cause this object to look like a data entity."""
        return self.id.__hash__()


class RemoteModelBase(ModelBase):

    # This should be set by the remote model during init time
    # The data() method will be using this
    _data_object = None

    def _data_item(self, data_object):
        data_fields = self._data_fields + self._auto_generated_attrs
        return {field: getattr(data_object, field) for field in data_fields}

    # data magic that will allow for a list of _data_object or a single item
    # if the object is a list, it will turn it into a list of hash's again
    def data(self, **options):
        if self._data_object is None:
            raise LookupError(_("data object is None"))
        if isinstance(self._data_object, list):
            return [self._data_item(item) for item in self._data_object]
        else:
            return self._data_item(self._data_object)


class NetworkRemoteModelBase(RemoteModelBase):

    network_driver = None

    @classmethod
    def get_driver(cls, context, region_name):
        if not cls.network_driver:
            cls.network_driver = import_class(CONF.network_driver)
        return cls.network_driver(context, region_name)


class NovaRemoteModelBase(RemoteModelBase):

    @classmethod
    def get_client(cls, context, region_name):
        return clients.create_nova_client(context, region_name)


class SwiftRemoteModelBase(RemoteModelBase):

    @classmethod
    def get_client(cls, context, region_name):
        return clients.create_swift_client(context, region_name)


class CinderRemoteModelBase(RemoteModelBase):

    @classmethod
    def get_client(cls, context):
        return clients.create_cinder_client(context)



# Copyright 2016 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from proboscis import SkipTest

from trove.tests.scenario.runners.test_runners import CheckInstance
from trove.tests.scenario.runners.test_runners import TestRunner


class InstanceErrorCreateRunner(TestRunner):

    def __init__(self):
        super(InstanceErrorCreateRunner, self).__init__(sleep_time=1)
        self.error_inst_id = None
        self.error2_inst_id = None

    def run_create_error_instance(self, expected_http_code=200):
        if self.is_using_existing_instance:
            raise SkipTest("Using an existing instance.")

        name = self.instance_info.name + '_error'
        flavor = self.get_instance_flavor(fault_num=1)

        client = self.auth_client
        inst = client.instances.create(
            name,
            self.get_flavor_href(flavor),
            self.instance_info.volume,
            nics=self.instance_info.nics,
            datastore=self.instance_info.dbaas_datastore,
            datastore_version=self.instance_info.dbaas_datastore_version)
        self.assert_client_code(client, expected_http_code)
        self.error_inst_id = inst.id

    def run_create_error2_instance(self, expected_http_code=200):
        if self.is_using_existing_instance:
            raise SkipTest("Using an existing instance.")

        name = self.instance_info.name + '_error2'
        flavor = self.get_instance_flavor(fault_num=2)

        client = self.auth_client
        inst = client.instances.create(
            name,
            self.get_flavor_href(flavor),
            self.instance_info.volume,
            nics=self.instance_info.nics,
            datastore=self.instance_info.dbaas_datastore,
            datastore_version=self.instance_info.dbaas_datastore_version)
        self.assert_client_code(client, expected_http_code)
        self.error2_inst_id = inst.id

    def run_wait_for_error_instances(self, expected_states=['ERROR']):
        error_ids = []
        if self.error_inst_id:
            error_ids.append(self.error_inst_id)
        if self.error2_inst_id:
            error_ids.append(self.error2_inst_id)

        if error_ids:
            self.assert_all_instance_states(
                error_ids, expected_states, fast_fail_status=[])

    def run_validate_error_instance(self):
        if not self.error_inst_id:
            raise SkipTest("No error instance created.")

        instance = self.get_instance(
            self.error_inst_id, self.auth_client)
        with CheckInstance(instance._info) as check:
            check.fault()

        err_msg = "disk is too small for requested image"
        self.assert_true(err_msg in instance.fault['message'],
                         "Message '%s' does not contain '%s'" %
                         (instance.fault['message'], err_msg))

    def run_validate_error2_instance(self):
        if not self.error2_inst_id:
            raise SkipTest("No error2 instance created.")

        instance = self.get_instance(
            self.error2_inst_id, client=self.admin_client)
        with CheckInstance(instance._info) as check:
            check.fault(is_admin=True)

        err_msg = "Quota exceeded for ram"
        self.assert_true(err_msg in instance.fault['message'],
                         "Message '%s' does not contain '%s'" %
                         (instance.fault['message'], err_msg))

    def run_delete_error_instances(self, expected_http_code=202):
        client = self.auth_client
        if self.error_inst_id:
            client.instances.delete(self.error_inst_id)
            self.assert_client_code(client, expected_http_code)
        if self.error2_inst_id:
            client.instances.delete(self.error2_inst_id)
            self.assert_client_code(client, expected_http_code)

    def run_wait_for_error_delete(self, expected_states=['SHUTDOWN']):
        delete_ids = []
        if self.error_inst_id:
            delete_ids.append(self.error_inst_id)
        if self.error2_inst_id:
            delete_ids.append(self.error2_inst_id)
        if delete_ids:
            self.assert_all_gone(delete_ids, expected_states[-1])
        else:
            raise SkipTest("Cleanup is not required.")

        # All the neutron ports should be removed.
        if self.error_inst_id:
            ports = self.neutron_client.list_ports(
                name='trove-%s' % self.error_inst_id
            )
            self.assert_equal(0, len(ports.get("ports", [])))



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""I totally stole most of this from melange, thx guys!!!"""

from collections import abc
import inspect
import os
import shlex
import shutil
import uuid
import urllib.parse as urlparse

from eventlet.timeout import Timeout
import jinja2
from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils.encodeutils import safe_encode
from oslo_utils import importutils
from oslo_utils import strutils
from passlib import pwd

from trove.common import cfg
from trove.common import exception
from trove.common.i18n import _


CONF = cfg.CONF
LOG = logging.getLogger(__name__)
import_class = importutils.import_class
import_object = importutils.import_object
import_module = importutils.import_module
bool_from_string = strutils.bool_from_string
execute = processutils.execute


def build_jinja_environment():
    env = jinja2.Environment(
        autoescape=True,
        loader=jinja2.ChoiceLoader([
            jinja2.FileSystemLoader(CONF.template_path),
            jinja2.PackageLoader("trove", "templates")
        ]))
    # Add some basic operation not built-in.
    env.globals['max'] = max
    env.globals['min'] = min
    return env


ENV = build_jinja_environment()


def pagination_limit(limit, default_limit):
    limit = int(limit or default_limit)
    return min(limit, default_limit)


def create_method_args_string(*args, **kwargs):
    """Returns a string representation of args and keyword args.

    I.e. for args=1,2,3 and kwargs={'a':4, 'b':5} you'd get: "1,2,3,a=4,b=5"
    """
    # While %s turns a var into a string but in some rare cases explicit
    # repr() is less likely to raise an exception.
    arg_strs = [repr(arg) for arg in args]
    arg_strs += ['%s=%s' % (repr(key), repr(value))
                 for (key, value) in kwargs.items()]
    return ', '.join(arg_strs)


def stringify_keys(dictionary):
    if dictionary is None:
        return None
    return {str(key): value for key, value in dictionary.items()}


def exclude(key_values, *exclude_keys):
    if key_values is None:
        return None
    return {key: value for key, value in key_values.items()
            if key not in exclude_keys}


def generate_uuid():
    return str(uuid.uuid4())


def raise_if_process_errored(process, exception):
    try:
        err = process.stderr.read()
        if err:
            raise exception(err)
    except OSError:
        pass


def clean_out(folder):
    for root, dirs, files in os.walk(folder):
        for f in files:
            os.unlink(os.path.join(root, f))
        for d in dirs:
            shutil.rmtree(os.path.join(root, d))


class cached_property(object):
    """A decorator that converts a function into a lazy property.

    Taken from : https://github.com/nshah/python-memoize
    The function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value:

        class Foo(object):

            @cached_property
            def bar(self):
                # calculate something important here
                return 42

    """

    def __init__(self, func, name=None, doc=None):
        self.func = func
        self.__name__ = name or func.__name__
        self.__doc__ = doc or func.__doc__

    def __get__(self, obj, owner):
        if obj is None:
            return self
        value = self.func(obj)
        setattr(obj, self.__name__, value)
        return value


class MethodInspector(object):

    def __init__(self, func):
        self._func = func

    @cached_property
    def required_args(self):
        return self.args[0:self.required_args_count]

    @cached_property
    def optional_args(self):
        keys = self.args[self.required_args_count: len(self.args)]
        return zip(keys, self.defaults)

    @cached_property
    def defaults(self):
        return self.argspec.defaults or ()

    @cached_property
    def required_args_count(self):
        return len(self.args) - len(self.defaults)

    @cached_property
    def args(self):
        args = self.argspec.args
        if inspect.ismethod(self._func):
            args.pop(0)
        return args

    @cached_property
    def argspec(self):
        return inspect.getargspec(self._func)

    def __str__(self):
        optionals = ["[{0}=<{0}>]".format(k) for k, v in self.optional_args]
        required = ["{0}=<{0}>".format(arg) for arg in self.required_args]
        args_str = ' '.join(required + optionals)
        return "%s %s" % (self._func.__name__, args_str)


def build_polling_task(retriever, condition=lambda value: value,
                       sleep_time=1, time_out=0, initial_delay=0):
    """Run a function in a loop with backoff on error.

    The condition function runs based on the retriever function result.
    """

    def poll_and_check():
        obj = retriever()
        if condition(obj):
            raise loopingcall.LoopingCallDone(retvalue=obj)

    call = loopingcall.BackOffLoopingCall(f=poll_and_check)
    return call.start(initial_delay=initial_delay,
                      starting_interval=sleep_time,
                      max_interval=30, timeout=time_out)


def wait_for_task(polling_task):
    """Waits for the task until it is finished"""
    try:
        return polling_task.wait()
    except loopingcall.LoopingCallTimeOut:
        raise exception.PollTimeOut


def poll_until(retriever, condition=lambda value: value,
               sleep_time=3, time_out=0, initial_delay=0):
    """Retrieves object until it passes condition, then returns it.

    If time_out_limit is passed in, PollTimeOut will be raised once that
    amount of time is eclipsed.

    """
    task = build_polling_task(retriever, condition=condition,
                              sleep_time=sleep_time, time_out=time_out,
                              initial_delay=initial_delay)
    return wait_for_task(task)


# Copied from nova.api.openstack.common in the old code.
def get_id_from_href(href):
    """Return the id or uuid portion of a url.

    Given: 'http://www.foo.com/bar/123?q=4'
    Returns: '123'

    Given: 'http://www.foo.com/bar/abc123?q=4'
    Returns: 'abc123'

    """
    return urlparse.urlsplit("%s" % href).path.split('/')[-1]


def execute_with_timeout(*args, **kwargs):
    time = kwargs.pop('timeout', CONF.command_process_timeout)
    log_output_on_error = kwargs.pop('log_output_on_error', False)

    timeout = Timeout(time)
    try:
        return execute(*args, **kwargs)
    except exception.ProcessExecutionError as e:
        if log_output_on_error:
            LOG.error(
                ("Command '%(cmd)s' failed. %(description)s "
                 "Exit code: %(exit_code)s\nstderr: %(stderr)s\n"
                 "stdout: %(stdout)s"),
                {'cmd': e.cmd, 'description': e.description or '',
                 'exit_code': e.exit_code, 'stderr': e.stderr,
                 'stdout': e.stdout})
        raise
    except Timeout as t:
        if t is not timeout:
            LOG.error("Got a timeout but not the one expected.")
            raise
        else:
            log_fmt = ("Time out after waiting "
                       "%(time)s seconds when running proc: %(args)s"
                       " %(kwargs)s.")
            exc_fmt = _("Time out after waiting "
                        "%(time)s seconds when running proc: %(args)s"
                        " %(kwargs)s.")
            msg_content = {
                'time': time, 'args': args,
                'kwargs': kwargs}
            LOG.error(log_fmt, msg_content)
            raise exception.ProcessExecutionError(exc_fmt % msg_content)
    finally:
        timeout.cancel()


def correct_id_with_req(id, request):
    # Due to a shortcoming with the way Trove uses routes.mapper,
    # URL entities right of the last slash that contain at least
    # one . are routed to our service without that suffix, as
    # it was interpreted as a filetype This method looks at the
    # request, and if applicable, reattaches the suffix to the id.
    routing_args = request.environ.get('wsgiorg.routing_args', [])
    for routing_arg in routing_args:
        try:
            found = routing_arg.get('format', '')
            if found and found not in CONF.expected_filetype_suffixes:
                return "%s.%s" % (id, found)
        except (AttributeError, KeyError):
            # Not the relevant routing_args entry.
            pass
    return id


def generate_random_password(password_length=None):
    password_length = (
        password_length or
        cfg.get_configuration_property('default_password_length')
    )
    return pwd.genword(length=password_length)


def try_recover(func):
    def _decorator(*args, **kwargs):
        recover_func = kwargs.pop("recover_func", None)
        try:
            func(*args, **kwargs)
        except Exception:
            if recover_func is not None:
                recover_func(func)
            else:
                LOG.debug("No recovery method defined for %(func)s", {
                          'func': func.__name__})
            raise
    return _decorator


def unpack_singleton(container):
    """Unpack singleton collections.

    Check whether a given collection is a singleton (has exactly one element)
    and unpack it if that is the case.
    Return the original collection otherwise.
    """
    if is_collection(container) and len(container) == 1:
        return unpack_singleton(container[0])

    return container


def is_collection(item):
    """Return True is a given item is an iterable collection, but not a string.
    """
    return (isinstance(item, abc.Iterable) and
            not isinstance(item, (bytes, str)))


def format_output(message, format_len=79, truncate_len=None, replace_index=0):
    """Recursive function to try and keep line lengths below a certain amount,
    so they can be displayed nicely on the command-line or UI.
    Tries replacement patterns one at a time (in round-robin fashion)
    that insert \n at strategic spots.
    """
    replacements = [['. ', '.\n'], [' (', '\n('], [': ', ':\n    ']]
    replace_index %= len(replacements)
    if not isinstance(message, list):
        message = message.splitlines(1)
    msg_list = []
    for line in message:
        if len(line) > format_len:
            ok_to_split_again = False
            for count in range(0, len(replacements)):
                lines = line.replace(
                    replacements[replace_index][0],
                    replacements[replace_index][1],
                    1
                ).splitlines(1)
                replace_index = (replace_index + 1) % len(replacements)
                if len(lines) > 1:
                    ok_to_split_again = True
                    break
            for item in lines:
                # If we spilt, but a line is still too long, do it again
                if ok_to_split_again and len(item) > format_len:
                    item = format_output(item, format_len=format_len,
                                         replace_index=replace_index)
                msg_list.append(item)
        else:
            msg_list.append(line)

    msg_str = "".join(msg_list)
    if truncate_len and len(msg_str) > truncate_len:
        msg_str = msg_str[:truncate_len - 3] + '...'
    return msg_str


def to_gb(bytes):
    """
    This was moved from dbaas.py so that it could be used as
    widely as a utility function. The tests corresponding to
    this were also moved out from test_dbaas.py to test_utils.py.
    """
    if bytes == 0:
        return 0.0
    size = bytes / 1024.0 ** 3
    # Make sure we don't return 0.0 if the size is greater than 0
    return max(round(size, 2), 0.01)


def to_mb(bytes):
    """
    This was moved from dbaas.py so that it could be used as
    widely as a utility function. The tests corresponding to
    this were also moved out from test_dbaas.py to test_utils.py.
    """
    if bytes == 0:
        return 0.0
    size = bytes / 1024.0 ** 2
    # Make sure we don't return 0.0 if the size is greater than 0
    return max(round(size, 2), 0.01)


def req_to_text(req):
    """
    We do a lot request logging for debug, but if the value of one
    request header is encoded in utf-8, an UnicodeEncodeError will
    be raised. So we should carefully encode request headers.

    To be consitent with webob, main procedures are copied from
    webob.Request.as_bytes.
    """
    url = req.url
    host = req.host_url
    assert url.startswith(host)
    url = url[len(host):]
    parts = [safe_encode('%s %s %s' % (req.method, url, req.http_version))]

    for k, v in sorted(req.headers.items()):
        header = safe_encode('%s: %s' % (k, v))
        parts.append(header)

    if req.body:
        parts.extend([b'', safe_encode(req.body)])

    return b'\r\n'.join(parts).decode(req.charset)


def validate_command(string):
    """
    Check if the string is legal for command

    raise invalidvalue if illegal
    """
    if string != shlex.quote(string):
        raise exception.InvalidValue(value=string)



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging

from trove.common.clients import create_nova_client


LOG = logging.getLogger(__name__)


class ServerGroup(object):

    @classmethod
    def load(cls, context, instance_id):
        client = create_nova_client(context)
        server_group = None
        expected_name = "locality_%s" % instance_id
        try:
            for sg in client.server_groups.list():
                if sg.name == expected_name:
                    server_group = sg
        except Exception:
            LOG.exception("Could not load server group for instance %s",
                          instance_id)

        if not server_group:
            LOG.info('No server group found for instance %s', instance_id)

        return server_group

    @classmethod
    def create(cls, context, locality, name_suffix):
        client = create_nova_client(context)
        server_group_name = "%s_%s" % ('locality', name_suffix)
        server_group = client.server_groups.create(
            name=server_group_name, policies=[locality])
        LOG.debug("Created '%(locality)s' server group called %(group_name)s "
                  "(id: %(group_id)s).",
                  {'locality': locality, 'group_name': server_group_name,
                   'group_id': server_group.id})

        return server_group

    @classmethod
    def delete(cls, context, server_group, force=False):
        # Only delete the server group if we're the last member in it, or if
        # it has no members
        if server_group:
            if force or len(server_group.members) <= 1:
                LOG.info("Deleting server group %s", server_group.id)
                client = create_nova_client(context)
                client.server_groups.delete(server_group.id)
            else:
                LOG.debug("Skipping delete of server group %(id)s "
                          "(members: %(members)s).",
                          {'id': server_group.id,
                           'members': server_group.members})

    @classmethod
    def convert_to_hint(cls, server_group, hints=None):
        if server_group:
            hints = hints or {}
            hints["group"] = server_group.id
        return hints

    @classmethod
    def build_scheduler_hint(cls, context, locality, name_suffix):
        scheduler_hint = None
        if locality:
            # Build the scheduler hint, but only if locality's a string
            if isinstance(locality, str):
                server_group = cls.create(
                    context, locality, name_suffix)
                scheduler_hint = cls.convert_to_hint(
                    server_group)
            else:
                # otherwise assume it's already in hint form (i.e. a dict)
                scheduler_hint = locality
        return scheduler_hint

    @classmethod
    def get_locality(cls, server_group):
        locality = None
        if server_group:
            locality = server_group.policies[0]
        return locality



# Copyright 2014 eBay Software Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_log import log as logging

from trove.cluster.service import ClusterController
import trove.common.apischema as apischema
from trove.common.auth import admin_context
from trove.common import exception
from trove.common.i18n import _
from trove.common import wsgi
from trove.extensions.mgmt.clusters import models
from trove.extensions.mgmt.clusters import views

LOG = logging.getLogger(__name__)


class MgmtClusterController(ClusterController):
    """Controller for cluster functionality."""
    schemas = apischema.mgmt_cluster

    @classmethod
    def get_action_schema(cls, body, action_schema):
        action_type = list(body.keys())[0]
        return action_schema.get(action_type, {})

    @admin_context
    def index(self, req, tenant_id):
        """Return a list of clusters."""
        LOG.debug("Showing a list of clusters for tenant '%s'.", tenant_id)
        LOG.info("req : '%s'\n\n", req)
        context = req.environ[wsgi.CONTEXT_KEY]
        deleted = None
        deleted_q = req.GET.get('deleted', '').lower()
        if deleted_q in ['true']:
            deleted = True
        elif deleted_q in ['false']:
            deleted = False
        clusters = models.MgmtCluster.load_all(context, deleted=deleted)
        view_cls = views.MgmtClustersView
        return wsgi.Result(view_cls(clusters, req=req).data(), 200)

    @admin_context
    def show(self, req, tenant_id, id):
        """Return a single cluster."""
        LOG.info("Showing cluster for tenant '%(tenant_id)s'.\n"
                 "req : '%(req)s'\n"
                 "id : '%(id)s'", {
                     "tenant_id": tenant_id, "req": req, "id": id})

        context = req.environ[wsgi.CONTEXT_KEY]
        cluster = models.MgmtCluster.load(context, id)
        return wsgi.Result(
            views.load_mgmt_view(cluster, req=req).data(),
            200)

    @admin_context
    def action(self, req, body, tenant_id, id):
        LOG.debug("Committing an action against cluster %(cluster)s for "
                  "tenant '%(tenant)s'.", {'cluster': id,
                                           'tenant': tenant_id})
        LOG.info("req : '%s'\n\n", req)
        if not body:
            raise exception.BadRequest(_("Invalid request body."))
        context = req.environ[wsgi.CONTEXT_KEY]
        cluster = models.MgmtCluster.load(context=context, id=id)

        if 'reset-task' in body:
            return self._action_reset_task(context, cluster, body)
        else:
            msg = _("Invalid cluster action requested.")
            raise exception.BadRequest(msg)

    def _action_reset_task(self, context, cluster, body):
        cluster.reset_task()
        return wsgi.Result(None, 202)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class QuotaView(object):

    def __init__(self, quotas):
        self.quotas = quotas

    def data(self):
        rtn = {}
        for resource_name, quota in self.quotas.items():
            rtn[resource_name] = quota.hard_limit
        return {'quotas': rtn}


class QuotaUsageView(object):

    def __init__(self, usages):
        self.usages = usages

    def data(self):
        return {'quotas': [{'resource': resource,
                            'in_use': usage['in_use'],
                            'reserved': usage['reserved'],
                            'limit': usage['limit']
                            } for resource, usage in self.usages.items()]}



#    Copyright 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import routes

from trove.backup.service import BackupController
from trove.backup.service import BackupStrategyController
from trove.cluster.service import ClusterController
from trove.common import wsgi
from trove.configuration.service import ConfigurationsController
from trove.configuration.service import ParametersController
from trove.datastore.service import DatastoreController
from trove.flavor.service import FlavorController
from trove.instance.service import InstanceController
from trove.limits.service import LimitsController
from trove.module.service import ModuleController
from trove.versions import VersionsController


class API(wsgi.Router):
    """Defines the API routes."""
    def __init__(self):
        mapper = routes.Mapper()
        super(API, self).__init__(mapper)
        self._instance_router(mapper)
        self._cluster_router(mapper)
        self._datastore_router(mapper)
        self._flavor_router(mapper)
        self._versions_router(mapper)
        self._limits_router(mapper)
        self._backups_router(mapper)
        self._backup_strategy_router(mapper)
        self._configurations_router(mapper)
        self._modules_router(mapper)

    def _versions_router(self, mapper):
        versions_resource = VersionsController().create_resource()
        mapper.connect("/",
                       controller=versions_resource,
                       action="show",
                       conditions={'method': ['GET']})

    def _datastore_router(self, mapper):
        datastore_resource = DatastoreController().create_resource()
        mapper.resource("datastore", "/{tenant_id}/datastores",
                        controller=datastore_resource)
        mapper.connect("/{tenant_id}/datastores/{datastore}/versions",
                       controller=datastore_resource,
                       action="version_index")
        mapper.connect("/{tenant_id}/datastores/{datastore}/versions/{id}",
                       controller=datastore_resource,
                       action="version_show")
        mapper.connect(
            "/{tenant_id}/datastores/{datastore}/versions/"
            "{version_id}/flavors",
            controller=datastore_resource,
            action="list_associated_flavors",
            conditions={'method': ['GET']}
        )
        mapper.connect(
            "/{tenant_id}/datastores/{datastore}/versions/"
            "{version_id}/volume-types",
            controller=datastore_resource,
            action="list_associated_volume_types",
            conditions={'method': ['GET']}
        )
        mapper.connect("/{tenant_id}/datastores/versions/{uuid}",
                       controller=datastore_resource,
                       action="version_show_by_uuid")

    def _instance_router(self, mapper):
        instance_resource = InstanceController().create_resource()
        mapper.connect("/{tenant_id}/instances",
                       controller=instance_resource,
                       action="index",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/instances/detail",
                       controller=instance_resource,
                       action="detail",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/instances",
                       controller=instance_resource,
                       action="create",
                       conditions={'method': ['POST']})
        mapper.connect("/{tenant_id}/instances/{id}",
                       controller=instance_resource,
                       action="show",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/instances/{id}/action",
                       controller=instance_resource,
                       action="action",
                       conditions={'method': ['POST']})
        mapper.connect("/{tenant_id}/instances/{id}",
                       controller=instance_resource,
                       action="update",
                       conditions={'method': ['PUT']})
        mapper.connect("/{tenant_id}/instances/{id}",
                       controller=instance_resource,
                       action="edit",
                       conditions={'method': ['PATCH']})
        mapper.connect("/{tenant_id}/instances/{id}",
                       controller=instance_resource,
                       action="delete",
                       conditions={'method': ['DELETE']})
        mapper.connect("/{tenant_id}/instances/{id}/backups",
                       controller=instance_resource,
                       action="backups",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/instances/{id}/configuration",
                       controller=instance_resource,
                       action="configuration",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/instances/{id}/log",
                       controller=instance_resource,
                       action="guest_log_list",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/instances/{id}/log",
                       controller=instance_resource,
                       action="guest_log_action",
                       conditions={'method': ['POST']})
        mapper.connect("/{tenant_id}/instances/{id}/modules",
                       controller=instance_resource,
                       action="module_list",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/instances/{id}/modules",
                       controller=instance_resource,
                       action="module_apply",
                       conditions={'method': ['POST']})
        mapper.connect("/{tenant_id}/instances/{id}/modules/{module_id}",
                       controller=instance_resource,
                       action="module_remove",
                       conditions={'method': ['DELETE']})

    def _cluster_router(self, mapper):
        cluster_resource = ClusterController().create_resource()
        mapper.connect("/{tenant_id}/clusters",
                       controller=cluster_resource,
                       action="index",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/clusters/{id}",
                       controller=cluster_resource,
                       action="show",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/clusters",
                       controller=cluster_resource,
                       action="create",
                       conditions={'method': ['POST']})
        mapper.connect("/{tenant_id}/clusters/{id}",
                       controller=cluster_resource,
                       action="action",
                       conditions={'method': ['POST']})
        mapper.connect("/{tenant_id}/clusters/{cluster_id}/instances/"
                       "{instance_id}",
                       controller=cluster_resource,
                       action="show_instance",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/clusters/{id}",
                       controller=cluster_resource,
                       action="delete",
                       conditions={'method': ['DELETE']})

    def _flavor_router(self, mapper):
        flavor_resource = FlavorController().create_resource()
        mapper.connect("/{tenant_id}/flavors",
                       controller=flavor_resource,
                       action="index",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/flavors/{id}",
                       controller=flavor_resource,
                       action="show",
                       conditions={'method': ['GET']})

    def _limits_router(self, mapper):
        limits_resource = LimitsController().create_resource()
        mapper.connect("/{tenant_id}/limits",
                       controller=limits_resource,
                       action="index",
                       conditions={'method': ['GET']})

    def _backups_router(self, mapper):
        backups_resource = BackupController().create_resource()
        mapper.connect("/{tenant_id}/backups",
                       controller=backups_resource,
                       action="index",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/backups",
                       controller=backups_resource,
                       action="create",
                       conditions={'method': ['POST']})
        mapper.connect("/{tenant_id}/backups/{id}",
                       controller=backups_resource,
                       action="show",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/backups/{id}",
                       controller=backups_resource,
                       action="delete",
                       conditions={'method': ['DELETE']})

    def _backup_strategy_router(self, mapper):
        backup_strategy_resource = BackupStrategyController().create_resource()
        mapper.connect("/{tenant_id}/backup_strategies",
                       controller=backup_strategy_resource,
                       action="create",
                       conditions={'method': ['POST']})
        mapper.connect("/{tenant_id}/backup_strategies",
                       controller=backup_strategy_resource,
                       action="index",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/backup_strategies",
                       controller=backup_strategy_resource,
                       action="delete",
                       conditions={'method': ['DELETE']})

    def _modules_router(self, mapper):

        modules_resource = ModuleController().create_resource()
        mapper.resource("modules", "/{tenant_id}/modules",
                        controller=modules_resource)
        mapper.connect("/{tenant_id}/modules",
                       controller=modules_resource,
                       action="index",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/modules",
                       controller=modules_resource,
                       action="create",
                       conditions={'method': ['POST']})
        mapper.connect("/{tenant_id}/modules/{id}",
                       controller=modules_resource,
                       action="show",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/modules/{id}",
                       controller=modules_resource,
                       action="update",
                       conditions={'method': ['PUT']})
        mapper.connect("/{tenant_id}/modules/{id}",
                       controller=modules_resource,
                       action="delete",
                       conditions={'method': ['DELETE']})
        mapper.connect("/{tenant_id}/modules/{id}/instances",
                       controller=modules_resource,
                       action="instances",
                       conditions={'method': ['GET']})
        mapper.connect("/{tenant_id}/modules/{id}/instances",
                       controller=modules_resource,
                       action="reapply",
                       conditions={'method': ['PUT']})

    def _configurations_router(self, mapper):
        parameters_resource = ParametersController().create_resource()
        path = '/{tenant_id}/datastores/versions/{version}/parameters'
        mapper.connect(path,
                       controller=parameters_resource,
                       action='index_by_version',
                       conditions={'method': ['GET']})
        path = '/{tenant_id}/datastores/versions/{version}/parameters/{name}'
        mapper.connect(path,
                       controller=parameters_resource,
                       action='show_by_version',
                       conditions={'method': ['GET']})

        path = '/{tenant_id}/datastores/{datastore}/versions/{id}'
        mapper.connect(path + '/parameters',
                       controller=parameters_resource,
                       action='index',
                       conditions={'method': ['GET']})
        mapper.connect(path + '/parameters/{name}',
                       controller=parameters_resource,
                       action='show',
                       conditions={'method': ['GET']})

        configuration_resource = ConfigurationsController().create_resource()
        mapper.connect('/{tenant_id}/configurations',
                       controller=configuration_resource,
                       action='index',
                       conditions={'method': ['GET']})
        mapper.connect('/{tenant_id}/configurations',
                       controller=configuration_resource,
                       action='create',
                       conditions={'method': ['POST']})
        mapper.connect('/{tenant_id}/configurations/{id}',
                       controller=configuration_resource,
                       action='show',
                       conditions={'method': ['GET']})
        mapper.connect('/{tenant_id}/configurations/{id}/instances',
                       controller=configuration_resource,
                       action='instances',
                       conditions={'method': ['GET']})
        mapper.connect('/{tenant_id}/configurations/{id}',
                       controller=configuration_resource,
                       action='edit',
                       conditions={'method': ['PATCH']})
        mapper.connect('/{tenant_id}/configurations/{id}',
                       controller=configuration_resource,
                       action='update',
                       conditions={'method': ['PUT']})
        mapper.connect('/{tenant_id}/configurations/{id}',
                       controller=configuration_resource,
                       action='delete',
                       conditions={'method': ['DELETE']})


def app_factory(global_conf, **local_conf):
    return API()



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import subprocess

from proboscis.asserts import fail
import tenacity

from trove import tests
from trove.tests.config import CONFIG
from trove.tests import util
from trove.tests.util.users import Requirements


def create_server_connection(instance_id, ip_address=None):
    if util.test_config.use_local_ovz:
        return OpenVZServerConnection(instance_id)
    return ServerSSHConnection(instance_id, ip_address=ip_address)


class ServerSSHConnection(object):
    def __init__(self, instance_id, ip_address=None):
        if not ip_address:
            req_admin = Requirements(is_admin=True)
            user = util.test_config.users.find_user(req_admin)
            dbaas_admin = util.create_dbaas_client(user)
            instance = dbaas_admin.management.show(instance_id)

            mgmt_interfaces = instance.server["addresses"].get(
                CONFIG.trove_mgmt_network, []
            )
            mgmt_addresses = [str(inf["addr"]) for inf in mgmt_interfaces
                              if inf["version"] == 4]

            if len(mgmt_addresses) == 0:
                fail("No IPV4 ip found for management network.")
            else:
                self.ip_address = mgmt_addresses[0]
        else:
            self.ip_address = ip_address

        TROVE_TEST_SSH_USER = os.environ.get('TROVE_TEST_SSH_USER')
        if TROVE_TEST_SSH_USER and '@' not in self.ip_address:
            self.ip_address = TROVE_TEST_SSH_USER + '@' + self.ip_address

    @tenacity.retry(
        wait=tenacity.wait_fixed(5),
        stop=tenacity.stop_after_attempt(3),
        retry=tenacity.retry_if_exception_type(subprocess.CalledProcessError)
    )
    def execute(self, cmd):
        exe_cmd = "%s %s '%s'" % (tests.SSH_CMD, self.ip_address, cmd)
        print("RUNNING COMMAND: %s" % exe_cmd)

        output = util.process(exe_cmd)

        print("OUTPUT: %s" % output)
        return output


class OpenVZServerConnection(object):
    def __init__(self, instance_id):
        self.instance_id = instance_id
        req_admin = Requirements(is_admin=True)
        self.user = util.test_config.users.find_user(req_admin)
        self.dbaas_admin = util.create_dbaas_client(self.user)
        self.instance = self.dbaas_admin.management.show(self.instance_id)
        self.instance_local_id = self.instance.server["local_id"]

    def execute(self, cmd):
        exe_cmd = "sudo vzctl exec %s %s" % (self.instance_local_id, cmd)
        print("RUNNING COMMAND: %s" % exe_cmd)
        return util.process(exe_cmd)



#    Copyright 2013 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from novaclient.v2.servers import Server
from proboscis import after_class
from proboscis import before_class
from proboscis import SkipTest
from proboscis import test
from proboscis.asserts import assert_equal
from proboscis.asserts import assert_raises

from trove.backup import models as backup_models
from trove.backup import state
from trove.common import exception
from trove.common.context import TroveContext
from trove.extensions.mgmt.instances.models import MgmtInstance
from trove.extensions.mgmt.instances.service import MgmtInstanceController
from trove.instance import models as imodels
from trove.instance import service_status as srvstatus
from trove.instance.models import DBInstance
from trove.instance.tasks import InstanceTasks
from trove.tests.config import CONFIG
from trove.tests.util import create_dbaas_client
from trove.tests.util import test_config
from trove.tests.util.users import Requirements

GROUP = "dbaas.api.mgmt.action.reset-task-status"


class MgmtInstanceBase(object):

    def setUp(self):
        self._create_instance()
        self.controller = MgmtInstanceController()

    def tearDown(self):
        self.db_info.delete()

    def _create_instance(self):
        self.context = TroveContext(is_admin=True)
        self.tenant_id = 999
        self.db_info = DBInstance.create(
            id="inst-id-1",
            name="instance",
            flavor_id=1,
            datastore_version_id=test_config.dbaas_datastore_version_id,
            tenant_id=self.tenant_id,
            volume_size=None,
            task_status=InstanceTasks.NONE)
        self.server = mock.MagicMock(spec=Server)
        self.instance = imodels.Instance(
            self.context,
            self.db_info,
            self.server,
            datastore_status=imodels.InstanceServiceStatus(
                srvstatus.ServiceStatuses.RUNNING))

    def _make_request(self, path='/', context=None, **kwargs):
        from webob import Request
        path = '/'
        print("path: %s" % path)
        return Request.blank(path=path, environ={'trove.context': context},
                             **kwargs)

    def _reload_db_info(self):
        self.db_info = DBInstance.find_by(id=self.db_info.id, deleted=False)


@test(groups=[GROUP])
class RestartTaskStatusTests(MgmtInstanceBase):

    @before_class
    def setUp(self):
        super(RestartTaskStatusTests, self).setUp()
        self.backups_to_clear = []

    @after_class
    def tearDown(self):
        super(RestartTaskStatusTests, self).tearDown()

    def _change_task_status_to(self, new_task_status):
        self.db_info.task_status = new_task_status
        self.db_info.save()

    def _make_request(self, path='/', context=None, **kwargs):
        req = super(RestartTaskStatusTests, self)._make_request(path, context,
                                                                **kwargs)
        req.method = 'POST'
        body = {'reset-task-status': {}}
        return req, body

    def reset_task_status(self):
        with mock.patch.object(MgmtInstance, 'load') as mock_load:
            mock_load.return_value = self.instance
            req, body = self._make_request(context=self.context)
            self.controller = MgmtInstanceController()
            resp = self.controller.action(req, body, self.tenant_id,
                                          self.db_info.id)

            mock_load.assert_called_once_with(context=self.context,
                                              id=self.db_info.id)
        return resp

    @test
    def mgmt_restart_task_requires_admin_account(self):
        context = TroveContext(is_admin=False)
        req, body = self._make_request(context=context)
        self.controller = MgmtInstanceController()
        assert_raises(exception.Forbidden, self.controller.action,
                      req, body, self.tenant_id, self.db_info.id)

    @test
    def mgmt_restart_task_returns_json(self):
        resp = self.reset_task_status()
        out = resp.data("application/json")
        assert_equal(out, None)

    @test
    def mgmt_restart_task_changes_status_to_none(self):
        self._change_task_status_to(InstanceTasks.BUILDING)
        self.reset_task_status()
        self._reload_db_info()
        assert_equal(self.db_info.task_status, InstanceTasks.NONE)

    @test
    def mgmt_reset_task_status_clears_backups(self):
        if CONFIG.fake_mode:
            raise SkipTest("Test requires an instance.")

        self.reset_task_status()
        self._reload_db_info()
        assert_equal(self.db_info.task_status, InstanceTasks.NONE)

        user = test_config.users.find_user(Requirements(is_admin=False))
        dbaas = create_dbaas_client(user)
        admin = test_config.users.find_user(Requirements(is_admin=True))
        admin_dbaas = create_dbaas_client(admin)
        result = dbaas.instances.backups(self.db_info.id)
        assert_equal(0, len(result))

        # Create some backups.
        backup_models.DBBackup.create(
            name="forever_new",
            description="forever new",
            tenant_id=self.tenant_id,
            state=state.BackupState.NEW,
            instance_id=self.db_info.id,
            deleted=False)

        backup_models.DBBackup.create(
            name="forever_build",
            description="forever build",
            tenant_id=self.tenant_id,
            state=state.BackupState.BUILDING,
            instance_id=self.db_info.id,
            deleted=False)

        backup_models.DBBackup.create(
            name="forever_completed",
            description="forever completed",
            tenant_id=self.tenant_id,
            state=state.BackupState.COMPLETED,
            instance_id=self.db_info.id,
            deleted=False)

        # List the backups for this instance.
        # There ought to be three in the admin tenant, but
        # none in a different user's tenant.
        result = dbaas.instances.backups(self.db_info.id)
        assert_equal(0, len(result))
        result = admin_dbaas.instances.backups(self.db_info.id)
        assert_equal(3, len(result))
        self.backups_to_clear = result

        # Reset the task status.
        self.reset_task_status()
        self._reload_db_info()
        result = admin_dbaas.instances.backups(self.db_info.id)
        assert_equal(3, len(result))
        for backup in result:
            if backup.name == 'forever_completed':
                assert_equal(backup.status,
                             state.BackupState.COMPLETED)
            else:
                assert_equal(backup.status, state.BackupState.FAILED)

    @test(runs_after=[mgmt_reset_task_status_clears_backups])
    def clear_test_backups(self):
        for backup in self.backups_to_clear:
            found_backup = backup_models.DBBackup.find_by(id=backup.id)
            found_backup.delete()
        admin = test_config.users.find_user(Requirements(is_admin=True))
        admin_dbaas = create_dbaas_client(admin)
        if not CONFIG.fake_mode:
            result = admin_dbaas.instances.backups(self.db_info.id)
            assert_equal(0, len(result))



# Copyright 2010-2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from trove.common import exception
from trove.common import policy
from trove.common import wsgi
from trove.flavor import models
from trove.flavor import views


class FlavorController(wsgi.Controller):
    """Controller for flavor functionality."""

    def show(self, req, tenant_id, id):
        """Return a single flavor."""
        context = req.environ[wsgi.CONTEXT_KEY]
        self._validate_flavor_id(id)
        flavor = models.Flavor(context=context, flavor_id=id)
        # Flavors do not bind to a particular tenant.
        # Only authorize the current tenant.
        policy.authorize_on_tenant(context, 'flavor:show')
        # Pass in the request to build accurate links.
        return wsgi.Result(views.FlavorView(flavor, req).data(), 200)

    def index(self, req, tenant_id):
        """Return all flavors."""
        context = req.environ[wsgi.CONTEXT_KEY]
        policy.authorize_on_tenant(context, 'flavor:index')
        flavors = models.Flavors(context=context)
        return wsgi.Result(views.FlavorsView(flavors, req).data(), 200)

    def _validate_flavor_id(self, id):
        if isinstance(id, str):
            return
        try:
            if int(id) != float(id):
                raise exception.NotFound(uuid=id)
        except ValueError:
            raise exception.NotFound(uuid=id)



# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from unittest.mock import Mock, patch
from testtools.matchers import Equals, Is, Not
import webob.exc

from trove.common import base_wsgi
from trove.common import exception
from trove.common import wsgi
from trove.tests.unittests import trove_testtools
import webob


class TestWsgi(trove_testtools.TestCase):
    def test_process_request(self):
        middleware = wsgi.ContextMiddleware("test_trove")
        req = webob.BaseRequest({})
        token = 'MI23fdf2defg123'
        user_id = 'test_user_id'
        req.headers = {
            'X-User': 'do not use - deprecated',
            'X-User-ID': user_id,
            'X-Auth-Token': token,
            'X-Service-Catalog': '[]'
        }
        req.environ = {}
        # invocation
        middleware.process_request(req)
        # assertions
        ctx = req.environ[wsgi.CONTEXT_KEY]
        self.assertThat(ctx, Not(Is(None)))
        self.assertThat(ctx.user_id, Equals(user_id))
        self.assertThat(ctx.auth_token, Equals(token))
        self.assertEqual(0, len(ctx.service_catalog))


class TestController(trove_testtools.TestCase):

    @patch.object(base_wsgi.Resource, 'execute_action',
                  side_effect=exception.RootHistoryNotFound())
    @patch.object(wsgi.Controller, 'delete', create=True)
    @patch.object(wsgi.Controller, 'validate_request')
    def test_exception_root_history_notfound(self, *args):
        controller = wsgi.Controller()
        resource = controller.create_resource()
        req = Mock()
        result = resource.execute_action('delete', req)
        self.assertIsInstance(result.wrapped_exc,
                              webob.exc.HTTPNotFound)



# Copyright [2015] Hewlett-Packard Development Company, L.P.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging

from trove.cluster import models
from trove.cluster.tasks import ClusterTasks
from trove.cluster.views import ClusterView
from trove.common import cfg
from trove.common import exception
from trove.common import server_group as srv_grp
from trove.common.strategies.cluster import base
from trove.common import utils
from trove.extensions.mgmt.clusters.views import MgmtClusterView
from trove.instance import models as inst_models
from trove.quota.quota import check_quotas
from trove.taskmanager import api as task_api


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class VerticaAPIStrategy(base.BaseAPIStrategy):

    @property
    def cluster_class(self):
        return VerticaCluster

    def _action_grow(self, cluster, body):
        nodes = body['grow']
        instances = []
        for node in nodes:
            instance = {
                'flavor_id': utils.get_id_from_href(node['flavorRef'])
            }
            if 'name' in node:
                instance['name'] = node['name']
            if 'volume' in node:
                instance['volume_size'] = int(node['volume']['size'])
            instances.append(instance)
        return cluster.grow(instances)

    def _action_shrink(self, cluster, body):
        nodes = body['shrink']
        instance_ids = [node['id'] for node in nodes]
        return cluster.shrink(instance_ids)

    @property
    def cluster_view_class(self):
        return VerticaClusterView

    @property
    def mgmt_cluster_view_class(self):
        return VerticaMgmtClusterView


class VerticaCluster(models.Cluster):

    @staticmethod
    def _create_instances(context, db_info, datastore, datastore_version,
                          instances, extended_properties, locality,
                          new_cluster=True):
        vertica_conf = CONF.get(datastore_version.manager)
        num_instances = len(instances)

        existing = inst_models.DBInstance.find_all(cluster_id=db_info.id,
                                                   deleted=False).all()
        num_existing = len(existing)

        # Matching number of instances with configured cluster_member_count
        if (new_cluster and
                num_instances != vertica_conf.cluster_member_count):
            raise exception.ClusterNumInstancesNotSupported(
                num_instances=vertica_conf.cluster_member_count)

        models.validate_instance_flavors(
            context, instances, vertica_conf.volume_support,
            vertica_conf.device_path)

        req_volume_size = models.get_required_volume_size(
            instances, vertica_conf.volume_support)
        models.assert_homogeneous_cluster(instances)

        models.validate_instance_nics(context, instances)

        deltas = {'instances': num_instances, 'volumes': req_volume_size}

        check_quotas(context.project_id, deltas)

        flavor_id = instances[0]['flavor_id']
        volume_size = instances[0].get('volume_size', None)
        volume_type = instances[0].get('volume_type', None)

        nics = [instance.get('nics', None) for instance in instances]

        azs = [instance.get('availability_zone', None)
               for instance in instances]

        regions = [instance.get('region_name', None)
                   for instance in instances]

        # Creating member instances
        minstances = []
        for i in range(0, num_instances):
            if i == 0 and new_cluster:
                member_config = {"id": db_info.id, "instance_type": "master"}
            else:
                member_config = {"id": db_info.id, "instance_type": "member"}
            instance_name = "%s-member-%s" % (db_info.name,
                                              str(i + num_existing + 1))
            minstances.append(
                inst_models.Instance.create(
                    context, instance_name, flavor_id,
                    datastore_version.image_id, [], [], datastore,
                    datastore_version, volume_size, None,
                    nics=nics[i], availability_zone=azs[i],
                    configuration_id=None, cluster_config=member_config,
                    volume_type=volume_type,
                    modules=instances[i].get('modules'), locality=locality,
                    region_name=regions[i])
            )
        return minstances

    @classmethod
    def create(cls, context, name, datastore, datastore_version,
               instances, extended_properties, locality, configuration,
               image_id=None):
        LOG.debug("Initiating cluster creation.")

        if configuration:
            raise exception.ConfigurationNotSupported()

        vertica_conf = CONF.get(datastore_version.manager)
        num_instances = len(instances)

        # Matching number of instances with configured cluster_member_count
        if num_instances != vertica_conf.cluster_member_count:
            raise exception.ClusterNumInstancesNotSupported(
                num_instances=vertica_conf.cluster_member_count)

        db_info = models.DBCluster.create(
            name=name, tenant_id=context.project_id,
            datastore_version_id=datastore_version.id,
            task_status=ClusterTasks.BUILDING_INITIAL)

        cls._create_instances(context, db_info, datastore, datastore_version,
                              instances, extended_properties, locality,
                              new_cluster=True)
        # Calling taskmanager to further proceed for cluster-configuration
        task_api.load(context, datastore_version.manager).create_cluster(
            db_info.id)

        return VerticaCluster(context, db_info, datastore, datastore_version)

    @staticmethod
    def k_safety(n):
        """
        Vertica defines k-safety values of 0, 1 or 2:
        https://my.vertica.com/docs/7.1.x/HTML/Content/Authoring/Glossary/
        K-Safety.htm
        """
        if n < 3:
            return 0
        elif n < 5:
            return 1
        else:
            return 2

    def grow(self, instances):
        LOG.debug("Growing cluster.")

        self.validate_cluster_available()

        context = self.context
        db_info = self.db_info
        datastore = self.ds
        datastore_version = self.ds_version

        db_info.update(task_status=ClusterTasks.GROWING_CLUSTER)

        locality = srv_grp.ServerGroup.convert_to_hint(self.server_group)
        new_instances = self._create_instances(context, db_info, datastore,
                                               datastore_version, instances,
                                               None, locality,
                                               new_cluster=False)

        task_api.load(context, datastore_version.manager).grow_cluster(
            db_info.id, [instance.id for instance in new_instances])

        return VerticaCluster(context, db_info, datastore, datastore_version)

    def shrink(self, instance_ids):
        self.validate_cluster_available()

        context = self.context
        db_info = self.db_info
        datastore_version = self.ds_version

        for db_instance in self.db_instances:
            if db_instance.type == 'master':
                if db_instance.id in instance_ids:
                    raise exception.ClusterShrinkInstanceInUse(
                        id=db_instance.id,
                        reason="Cannot remove master node."
                    )

        all_instance_ids = [db_instance.id for db_instance
                            in self.db_instances]

        left_instances = [instance_id for instance_id
                          in all_instance_ids
                          if instance_id not in instance_ids]

        k = self.k_safety(len(left_instances))

        vertica_conf = CONF.get(datastore_version.manager)
        if k < vertica_conf.min_ksafety:
            raise exception.ClusterNumInstancesBelowSafetyThreshold()

        db_info.update(task_status=ClusterTasks.SHRINKING_CLUSTER)

        task_api.load(context, datastore_version.manager).shrink_cluster(
            self.db_info.id, instance_ids)
        return VerticaCluster(self.context, db_info,
                              self.ds, self.ds_version)


class VerticaClusterView(ClusterView):

    def build_instances(self):
        return self._build_instances(['member', 'master'],
                                     ['member', 'master'])


class VerticaMgmtClusterView(MgmtClusterView):

    def build_instances(self):
        return self._build_instances(['member', 'master'],
                                     ['member', 'master'])



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from trove.common import cfg
from trove.common.i18n import _
from trove.common import utils

CONF = cfg.CONF

"""
The classes below are generic and can be used for any datastore, but will not
provide validation. To add a new datastore create a sub-package (see mysql for
example) and create new child classes inheriting from these generic classes.

As a guideline, for new datastores the following class methods/variables should
be overridden if validation is desired (see their docstrings for additional
info):

DatastoreModelsBase:
        __init__

DatastoreSchema:
        _max_schema_name_length
        _is_valid_schema_name
        verify_dict
        _create_checks
        _delete_checks

DatastoreUser:
        _is_valid_user_name
        _is_valid_host_name
        _is_valid_password
        _is_valid_database
        verify_dict
        _create_checks
        _delete_checks
"""


class DatastoreModelsBase(object):
    """Base model for the datastore schema and user models."""

    def serialize(self):
        return self.__dict__

    def _deserialize(self, obj):
        self.__dict__ = obj

    def __repr__(self):
        return str(self.serialize())

    @classmethod
    def deserialize(cls, value, verify=True):
        item = cls(deserializing=True)
        item._deserialize(value)
        if verify:
            item.verify_dict()
        return item

    @abc.abstractmethod
    def verify_dict(self):
        """Validate the object's data dictionary.
        :returns:            True if dictionary is valid.
        """

    @staticmethod
    def check_string(value, desc):
        """Check if the value is a string/unicode.
        :param value:        Value to check.
        :param desc:         Description for exception message.
        :raises:             ValueError if not a string/unicode.
        """
        if not isinstance(value, str):
            raise ValueError(_("%(desc)s is not a string. Type = %(t)s.")
                             % {'desc': desc, 't': type(value)})


class DatastoreSchema(DatastoreModelsBase):
    """Represents a database schema."""

    def __init__(self, name=None, deserializing=False):
        self._name = None
        self._collate = None
        self._character_set = None
        # If both or neither are passed in this is a bug.
        if bool(deserializing) == bool(name):
            raise RuntimeError(_("Bug in DatastoreSchema()"))
        if not deserializing:
            self.name = name

    def __str__(self):
        return str(self.name)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._validate_schema_name(value)
        self._name = value

    def _validate_schema_name(self, value):
        """Perform checks on a given schema name.
        :param value:        Validated schema name.
        :type value:         string
        :raises:             ValueError On validation errors.
        """
        if not value:
            raise ValueError(_("Schema name empty."))
        self.check_string(value, 'Schema name')
        if self._max_schema_name_length and (len(value) >
                                             self._max_schema_name_length):
            raise ValueError(_("Schema name '%(name)s' is too long. "
                               "Max length = %(max_length)d.")
                             % {'name': value,
                                'max_length': self._max_schema_name_length})
        elif not self._is_valid_schema_name(value):
            raise ValueError(_("'%s' is not a valid schema name.") % value)

    @property
    def _max_schema_name_length(self):
        """Return the maximum valid schema name length if any.
        :returns:            Maximum schema name length or None if unlimited.
        """
        return None

    def _is_valid_schema_name(self, value):
        """Validate a given schema name.
        :param value:        Validated schema name.
        :type value:         string
        :returns:            TRUE if valid, FALSE otherwise.
        """
        return True

    def verify_dict(self):
        """Check that the object's dictionary values are valid by reloading
        them via the property setters. The checkers should raise the
        ValueError exception if invalid. All mandatory fields should be
        checked.
        """
        self.name = self._name

    @property
    def ignored_dbs(self):
        return cfg.get_ignored_dbs()

    def is_ignored(self):
        return self.name in self.ignored_dbs

    def check_reserved(self):
        """Check if the name is on the ignore_dbs list, meaning it is
        reserved.
        :raises:             ValueError if name is on the reserved list.
        """
        if self.is_ignored():
            raise ValueError(_('Database name "%(name)s" is on the reserved '
                               'list: %(reserved)s.')
                             % {'name': self.name,
                                'reserved': self.ignored_dbs})

    def _create_checks(self):
        """Checks to be performed before database can be created."""
        self.check_reserved()

    def check_create(self):
        """Check if the database can be created.
        :raises:             ValueError if the schema is not valid for create.
        """
        try:
            self._create_checks()
        except ValueError as e:
            raise ValueError(_('Cannot create database: %(error)s')
                             % {'error': str(e)})

    def _delete_checks(self):
        """Checks to be performed before database can be deleted."""
        self.check_reserved()

    def check_delete(self):
        """Check if the database can be deleted.
        :raises:             ValueError if the schema is not valid for delete.
        """
        try:
            self._delete_checks()
        except ValueError as e:
            raise ValueError(_('Cannot delete database: %(error)s')
                             % {'error': str(e)})


class DatastoreUser(DatastoreModelsBase):
    """Represents a datastore user."""

    _HOSTNAME_WILDCARD = '%'
    root_username = 'root'

    def __init__(self, name=None, password=None, host=None, databases=None,
                 deserializing=False):
        self._name = None
        self._password = None
        self._host = self._HOSTNAME_WILDCARD
        self._databases = []
        self._is_root = False
        if not deserializing:
            self.name = name
            if password:
                self.password = password
            if host:
                self.host = host
            if databases:
                self.databases = databases

    @classmethod
    def root(cls, name=None, password=None, *args, **kwargs):
        if not name:
            name = cls.root_username
        if not password:
            password = utils.generate_random_password()
        user = cls(name, password, *args, **kwargs)
        user.make_root()
        return user

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._validate_user_name(value)
        self._name = value

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, value):
        self.check_string(value, "User password")
        if self._is_valid_password(value):
            self._password = value
        else:
            raise ValueError(_("'%s' is not a valid password.") % value)

    def _add_database(self, value):
        serial_db = self._build_database_schema(value).serialize()
        if self._is_valid_database(serial_db):
            self._databases.append(serial_db)

    @property
    def databases(self):
        return self._databases

    @databases.setter
    def databases(self, value):
        if isinstance(value, list):
            for dbname in value:
                self._add_database(dbname)
        else:
            self._add_database(value)

    @property
    def host(self):
        if self._host is None:
            return self._HOSTNAME_WILDCARD
        return self._host

    @host.setter
    def host(self, value):
        self.check_string(value, "User host name")
        if self._is_valid_host_name(value):
            self._host = value
        else:
            raise ValueError(_("'%s' is not a valid hostname.") % value)

    def _build_database_schema(self, name):
        """Build a schema for this user.
        :type name:             string
        """
        return self.schema_model(name)

    def deserialize_schema(self, value):
        """Deserialize a user's databases value.
        :type value:            dict
        """
        return self.schema_model.deserialize(value)

    def _validate_user_name(self, value):
        """Perform validations on a given user name.
        :param value:        Validated user name.
        :type value:         string
        :raises:             ValueError On validation errors.
        """
        if not value:
            raise ValueError(_("User name empty."))
        self.check_string(value, "User name")
        if self._max_user_name_length and (len(value) >
                                           self._max_user_name_length):
            raise ValueError(_("User name '%(name)s' is too long. "
                               "Max length = %(max_length)d.")
                             % {'name': value,
                                'max_length': self._max_user_name_length})
        elif not self._is_valid_user_name(value):
            raise ValueError(_("'%s' is not a valid user name.") % value)

    @property
    def _max_user_name_length(self):
        """Return the maximum valid user name length if any.
        :returns:            Maximum user name length or None if unlimited.
        """
        return None

    def _is_valid_user_name(self, value):
        """Validate a given user name.
        :param value:        User name to be validated.
        :type value:         string
        :returns:            TRUE if valid, FALSE otherwise.
        """
        return True

    def _is_valid_host_name(self, value):
        """Validate a given host name.
        :param value:        Host name to be validated.
        :type value:         string
        :returns:            TRUE if valid, FALSE otherwise.
        """
        return True

    def _is_valid_password(self, value):
        """Validate a given password.
        :param value:        Password to be validated.
        :type value:         string
        :returns:            TRUE if valid, FALSE otherwise.
        """
        return True

    def _is_valid_database(self, value):
        """Validate a given database (serialized schema object).
        :param value:        The database to be validated.
        :type value:         dict
        :returns:            TRUE if valid, FALSE otherwise.
        :raises:             ValueError if operation not allowed.
        """
        return value not in self.databases

    def verify_dict(self):
        """Check that the object's dictionary values are valid by reloading
        them via the property setters. The checkers should raise the
        ValueError exception if invalid. All mandatory fields should be
        checked.
        """
        self.name = self._name
        if self.__dict__.get('_password'):
            self.password = self._password
        else:
            self._password = None
        if self.__dict__.get('_host'):
            self.host = self._host
        else:
            self._host = self._HOSTNAME_WILDCARD
        if self.__dict__.get('_databases'):
            for database in self._databases:
                # Create the schema for validation only
                self.deserialize_schema(database)
        else:
            self._databases = []
        if not self.__dict__.get('_is_root'):
            self._is_root = False

    @property
    def schema_model(self):
        return DatastoreSchema

    @property
    def ignored_users(self):
        if self._is_root:
            return []
        return cfg.get_ignored_users()

    @property
    def is_ignored(self):
        return self.name in self.ignored_users

    def make_root(self):
        self._is_root = True

    def check_reserved(self):
        """Check if the name is on the ignore_users list, meaning it is
        reserved.
        :raises:             ValueError if name is on the reserved list.
        """
        if self.is_ignored:
            raise ValueError(_('User name "%(name)s" is on the reserved '
                               'list: %(reserved)s.')
                             % {'name': self.name,
                                'reserved': self.ignored_users})

    def _create_checks(self):
        """Checks to be performed before user can be created."""
        self.check_reserved()

    def check_create(self):
        """Check if the user can be created.
        :raises:             ValueError if the user is not valid for create.
        """
        try:
            self._create_checks()
        except ValueError as e:
            raise ValueError(_('Cannot create user: %(error)s')
                             % {'error': str(e)})

    def _delete_checks(self):
        """Checks to be performed before user can be created."""
        self.check_reserved()

    def check_delete(self):
        """Check if the user can be deleted.
        :raises:             ValueError if the user is not valid for delete.
        """
        try:
            self._delete_checks()
        except ValueError as e:
            raise ValueError(_('Cannot delete user: %(error)s')
                             % {'error': str(e)})
